// Copyright (c) 2014-2020 NetEase, Inc.
// All right reserved.

part of room_kit;

///房间内音频控制器.
class _NEInRoomAudioControllerImpl extends NEInRoomAudioController
    with _RoomAloggerMixin {

  final _InRoomData roomData;

  final NEInRoomService inRoomService;

  _NEInRoomAudioControllerImpl(this.roomData, this.inRoomService);

  Future<void> setUp() async {
    await syncLocalAudioStatus();
  }

  @override
  Future<NEResult> syncLocalAudioStatus() async {
    commonLogger.i('syncLocalAudioStatus: ${roomData.self.audioStatus}');
    var status = await Permission.microphone.status;
    var permissionGranted = status == PermissionStatus.granted;
    if (!permissionGranted) {
      commonLogger.i('syncLocalAudioStatus: no permission');
      return  NEResult(code:RoomErrorCode.failed);
    }
    var result = await NERoomMediaService.instance.engine.enableLocalAudio(
        roomData.self.audioStatus == NERoomAudioStatus.on);
    commonLogger.i('syncLocalAudioStatus: result $result');
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> adjustPlaybackSignalVolume(int volume) async {
    apiLogger.i('adjustPlaybackSignalVolume: $volume');
    final result = await NERoomMediaService.instance.engine
        .adjustPlaybackSignalVolume(volume);
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> adjustRecordingSignalVolume(int volume) async {
    apiLogger.i('adjustRecordingSignalVolume: $volume');
    final result = await NERoomMediaService.instance.engine
        .adjustRecordingSignalVolume(volume);
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> enableAudioVolumeIndication(
      bool enable, int interval) async {
    apiLogger.i('enableAudioVolumeIndication: $enable $interval');
    final result = await NERoomMediaService.instance.engine
        .enableAudioVolumeIndication(enable, interval);
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> enableEarBack(bool enable, int volume) async {
    apiLogger.i('enableEarBack: $enable $volume');
    final result = await NERoomMediaService.instance.engine.deviceManager
        .enableEarBack(enable, volume);
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  // @override
  // Future<NEResult<void>> enableMyAudio(bool enable) async {
  //   apiLogger.i('enableMyAudio: $enable');
  //   final result =
  //       await NERoomMediaService.instance.engine.enableLocalAudio(enable);
  //   return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  // }

  @override
  bool isMuteOnEntryOn() {
    return roomData.roomInfo?.audioAllMute == true
        || roomData.roomInfo?.settings?.isAttendeeAudioOff == true;
  }

  @override
  Future<bool> isMyLoudSpeakerOn() {
    return NERoomMediaService.instance.engine.deviceManager.isSpeakerphoneOn();
  }

  @override
  Future<bool> isPlayoutDeviceMute() {
    return NERoomMediaService.instance.engine.deviceManager
        .isPlayoutDeviceMute();
  }

  @override
  Future<bool> isRecordDeviceMute() {
    return NERoomMediaService.instance.engine.deviceManager
        .isRecordDeviceMute();
  }

  @override
  Future<NEResult<void>> muteAllParticipantsAudio(bool allowUnmuteSelf) {
    apiLogger.i('muteAllParticipantsAudio: allowUnmuteSelf=$allowUnmuteSelf');
    if (allowUnmuteSelf) {
      return InRoomRepository.hostMuteAllAudio(roomData.roomId);
    } else {
      return InRoomRepository.hostMuteAllAudioOff(roomData.roomId);
    }
  }

  @override
  Future<NEResult<void>> unmuteAllParticipantsAudio() {
    apiLogger.i('unmuteAllParticipantsAudio');
    return InRoomRepository.hostUnMuteAllAudio(roomData.roomId);
  }

  @override
  Future<NEResult<void>> muteMyAudio(bool isMute) {
    apiLogger.i('muteMyAudio: mute=$isMute');
    return isMute
        ? InRoomRepository.muteSelfAudio(roomData.roomId)
        : InRoomRepository.unMuteSelfAudio(roomData.roomId);
  }

  @override
  bool canUnmuteMyAudio() {
    return inRoomService.isMySelfHost() ||
        inRoomService.isParticipantsUnmuteSelfAllowed() ||
        roomData.self.audioStatus == NERoomAudioStatus.waitingOpen;
  }

  @override
  Future<NEResult<void>> muteParticipantAudio(String userId, bool isMute) {
    apiLogger.i('muteParticipantAudio: userId=$userId mute=$isMute');
    if (roomData.myUserId == userId) {
      return isMute
          ? InRoomRepository.muteSelfAudio(roomData.roomId)
          : InRoomRepository.unMuteSelfAudio(roomData.roomId);
    } else {
      return isMute
          ? InRoomRepository.hostMuteAudio(roomData.roomId, [userId])
          : InRoomRepository.hostUnMuteAudio(roomData.roomId, [userId]);
    }
  }

  @override
  Future<NEResult<void>> lowerAllHands() {
    apiLogger.i('lowerAllHands');
    final handsUpUserList = roomData.userId2User.values
        .where((user) =>
            user.isOnline && user.raiseHandDetail.status == NEHandsUpStatus.up)
        .map((user) => user.userId)
        .toList(growable: false);
    return InRoomRepository.hostRejectAudioHandsUp(
        roomData.roomId, handsUpUserList);
  }

  @override
  Future<NEResult<void>> lowerHand(String userId) {
    apiLogger.i('lowerHand: userId=$userId ${roomData.getUserName(userId)}');
    return InRoomRepository.hostRejectAudioHandsUp(roomData.roomId, [userId]);
  }

  @override
  Future<NEResult<void>> raiseMyHand() {
    apiLogger.i('raiseMyHand');
    return InRoomRepository.muteAllHandsUp(roomData.roomId);
  }

  @override
  Future<NEResult<void>> lowerMyHand() {
    apiLogger.i('lowerMyHand');
    return InRoomRepository.muteAllUnHandsUp(roomData.roomId);
  }

  @override
  Future<NEResult<void>> setEarBackVolume(int volume) async {
    apiLogger.i('setEarBackVolume: $volume');
    final result = await NERoomMediaService.instance.engine.deviceManager
        .setEarBackVolume(volume);
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> setMyLoudSpeakerOn(bool on) async {
    apiLogger.i('setMyLoudSpeakerOn: $on');
    final result = await NERoomMediaService.instance.engine.deviceManager
        .setSpeakerphoneOn(on);
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> setPlayoutDeviceMute(bool isMute) async {
    apiLogger.i('setPlayoutDeviceMute: $isMute');
    final result = await NERoomMediaService.instance.engine.deviceManager
        .setPlayoutDeviceMute(isMute);
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> setRecordDeviceMute(bool isMute) async {
    apiLogger.i('setRecordDeviceMute: $isMute');
    final result = await NERoomMediaService.instance.engine.deviceManager
        .setRecordDeviceMute(isMute);
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> startAudioDump() async {
    apiLogger.i('startAudioDump');
    final result = await NERoomMediaService.instance.engine.startAudioDump();
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> stopAudioDump() async {
    apiLogger.i('stopAudioDump');
    final result = await NERoomMediaService.instance.engine.stopAudioDump();
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> subscribeRemoteAudioStream(String userId) async {
    final result = await toggleSubscribeRemoteAudioStream(userId, true);
    apiLogger.i('subscribeRemoteAudioStream: $userId $result');
    return NEResult(code: result);
  }

  @override
  Future<NEResult<void>> unsubscribeRemoteAudioStream(String userId) async {
    final result = await toggleSubscribeRemoteAudioStream(userId, false);
    apiLogger.i('unsubscribeRemoteAudioStream: $userId $result');
    return NEResult(code: result);
  }

  Future<int> toggleSubscribeRemoteAudioStream(
      String userId, bool subscribe) async {
    if (userId == roomData.myUserId) {
      return RoomErrorCode.paramsError;
    }
    final rtcId = roomData.userId2User[userId]?.rtcId;
    var result = RoomErrorCode.paramsError;
    if (rtcId != null) {
      result = await NERoomMediaService.instance.engine
          .subscribeRemoteAudio(rtcId, subscribe);
    }
    return result == 0 ? RoomErrorCode.success : result;
  }

  @override
  Future<NEResult<void>> subscribeAllRemoteAudioStream() async {
    apiLogger.i('subscribeAllRemoteAudioStream');
    final result = await NERoomMediaService.instance.engine.subscribeAllRemoteAudio(true);
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> unsubscribeAllRemoteAudioStream() async {
    apiLogger.i('unsubscribeAllRemoteAudioStream');
    final result = await NERoomMediaService.instance.engine.subscribeAllRemoteAudio(false);
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }
}
