// Copyright (c); 2014-2020 NetEase, Inc.
// All right reserved.
part of room_kit;

/// 在房间里获取聊天室控制器实例.
class _NEInRoomChatControllerImpl extends NEInRoomChatController with _RoomAloggerMixin{

  final _InRoomData roomData;
  bool hasEnter = false;
  _NEInRoomChatControllerImpl(this.roomData, List<NEInRoomServiceListener> listeners,){
     ChatRoomRepository.chatRoomReceiveMessage.listen((event) {
       listeners.forEach((element) {
         element.onRoomChatMessageReceived(mapReduce(event));
       });
     }) ;
  }

  @override
  Future<NEResult<void>> enterChatRoom() async {
    apiLogger.i('enterChatRoom');
    var roomId = roomData.joinRoomInfo.chatRoomId;

    ///判断是否支持
    if (!isChatroomEnabled() || roomId == null) {
      return NEResult(code: RoomErrorCode.failed);
    }

    ///如果已经进入，则直接返回
    if (hasEnter) return NEResult.success();

    var data = EnterChatRoomData();
    data.roomId = roomId;
    data.nick = UserProfile.displayName;
    data.account = UserProfile.imAccid;
    data.token = UserProfile.imToken;
    var result = await ChatRoomRepository.enterChatRoom(data, 1,
        linkAddress: (String? roomId, String? account) async {
      var result = await ChatRoomRepository.fetchRoomLink(roomId!);
      return (result.code == RoomErrorCode.success && result.data is List)
          ? result.data
          : null;
    });
    hasEnter = result.isSuccess();
    return result;
  }

  @override
  Future<NEResult<void>> exitChatRoom() async {
    apiLogger.i('exitChatRoom');
    ///如果已经退出，则直接退出
    if (!hasEnter) return NEResult.success();
    var roomId = roomData.joinRoomInfo.chatRoomId;
    NEResult<void> result;
    if (roomId != null) {
      result = await ChatRoomRepository.exitChatRoom(roomId);
      hasEnter = !result.isSuccess();
    } else {
      result = NEResult(code: RoomErrorCode.failed);
    }
    return result;
  }

  @override
  bool isChatroomEnabled() {
    return NERoomKit.instance.getSettingsService().isChatroomSupported()
    && TextUtils.isNotEmpty(roomData.joinRoomInfo.chatRoomId);
  }

  @override
  bool hasEnterChatRoom() => hasEnter;

  @override
  Future<NEResult<NEChatRoomMessage>> sendChatRoomMessage(NEChatRoomMessage message) async {
    apiLogger.i('sendChatRoomMessage ${message} ');
    var roomId = roomData.joinRoomInfo.chatRoomId;
    ///判断是否支持
    if (!isChatroomEnabled() ||
        roomId == null ||
        !hasEnter) {
      return NEResult(code: RoomErrorCode.failed);
    }

    message.time = DateTime.now().millisecondsSinceEpoch;
    message._chatRoomUserInfo = NEChatRoomUserInfo(
        nick: UserProfile.displayName, isSelf: true);
    var chatRoomMessage = ChatRoomMessage();
    var extension = ChatRoomMessageExtension();
    extension.nick = message.chatRoomUserInfo.nick;
    extension.senderExtension = message.senderExtension;


    if (message is NEChatRoomTextMessage) {
      if (TextUtils.isEmpty(message.content)) {
        return NEResult(code: RoomErrorCode.failed);
      }
      chatRoomMessage =
          ChatRoomMessageBuilder.createTextMessage(roomId, message.content!);
    } else if (message is NEChatRoomImageMessage) {
      chatRoomMessage =
          ChatRoomMessageBuilder.createImageMessage(
              roomId, message.attachment.url, message.attachment.displayName);

      ///TODO 返回message._attachment的相关属性
      ///   message._attachment.size = ;
    } else if (message is NEChatRoomFileMessage) {
      chatRoomMessage =
          ChatRoomMessageBuilder.createFileMessage(
              roomId, message.attachment.url, message.attachment.displayName);
    }
    chatRoomMessage
      ..time = message.time
      ..fromAccount = UserProfile.imAccid
      ..messageExtension = extension;
    var result = await ChatRoomRepository.sendChatRoomMessage(chatRoomMessage);
    ///处理返回消息结构体的发送状态
    message._msgStatus =
        result.isSuccess() ? MsgStatus.success : MsgStatus.fail;
    return NEResult(code: result.code, data: message);
  }

  @override
  Future<NEResult<void>> updateChatRoomUserRole(
      NEChatRoomUserRole chatRoomUserRole,
      {Map<String, dynamic>? extension}) async {
    apiLogger.i('chatRoomUserInfo ${extension}');
    if (!hasEnter) {
      return NEResult(code: RoomErrorCode.failed);
    } else {
      return await ChatRoomRepository.updateChatRoomRole(
          roomData.joinRoomInfo.chatRoomId!,
          chatRoomUserRole.nick!,
          chatRoomUserRole.avatar,
          null,
          true);
    }
  }

  List<NEChatRoomMessage> mapReduce(List<ChatRoomMessage> event) {
    var result = <NEChatRoomMessage>[];
    event.forEach((ChatRoomMessage message) {
      if (message.sessionId == roomData.joinRoomInfo.chatRoomId) {
        var neChatMsg = mapChatMessage(message);
        if (neChatMsg != null) result.add(neChatMsg);
      }
    });
    commonLogger.i('chatRoomReceiveMessage mapReduce result,${result}');
    return result;
  }

  NEChatRoomMessage? mapChatMessage(ChatRoomMessage message) {
    NEChatRoomMessage? chatMessage;
    var chatRoomUserInfo = NEChatRoomUserInfo(
        isSelf: message.fromAccount == UserProfile.imAccid,
        avatar: message.messageExtension!.avatar,
        nick: message.messageExtension!.nick);
    switch (message.msgType) {
      case MsgType.text:
        chatMessage = NEChatRoomTextMessage(
          senderExtension: message.messageExtension!.senderExtension,
          content: message.content,
          time: message.time,
        );
        break;
      case MsgType.image:
        var attachment = ImageAttachment();
        if (message.attach != null) {
          attachment.fromJson(json.decode(message.attach!) as Map);
          chatMessage = NEChatRoomImageMessage(
            senderExtension: message.messageExtension!.senderExtension,
            content: message.content,
            time: message.time,
            displayName: attachment.displayName!,
            path: attachment.url!,
            width: attachment.width,
            height: attachment.height,
          );
          (chatMessage as NEChatRoomImageMessage)._attachment
            ..md5 = attachment.md5
            ..size = attachment.size
            ..extension = attachment.extension
            ..expire = attachment.expire;
        }
        break;
      case MsgType.file:
        var attachment = FileAttachment();
        if (message.attach != null) {
          attachment.fromJson(json.decode(message.attach!) as Map);
          chatMessage = NEChatRoomFileMessage(
            senderExtension: message.messageExtension!.senderExtension,
            content: message.content,
            time: message.time,
            displayName: attachment.displayName!,
            path: attachment.url!,
          );
          (chatMessage as NEChatRoomFileMessage)._attachment
            ..md5 = attachment.md5
            ..size = attachment.size
            ..extension = attachment.extension
            ..expire = attachment.expire;
        }
        break;
    }
    if (chatMessage != null) {
      chatMessage._msgStatus = MsgStatus.success;
      chatMessage._chatRoomUserInfo = chatRoomUserInfo;
    }
    return chatMessage;
  }

  void release() {
    exitChatRoom();
  }
}
