// Copyright (c) 2014-2021 NetEase, Inc.
// All right reserved.
part of room_kit;

/// 屏幕共享控制器
class _NEInRoomScreenShareControllerImpl extends NEInRoomScreenShareController
    with _RoomAloggerMixin, NERtcScreenShareBroadcastCallback {

  final _InRoomData roomData;

  NERtcScreenShare? iosScreenSharePlugin;

  Completer<NEResult<void>>? startScreenShareCompleter;

  _NEInRoomScreenShareControllerImpl(this.roomData);

  @override
  Future<NEResult<void>> stopParticipantScreenShare(String userId) {
    apiLogger.i('stopParticipantScreenShare: $userId');
    return InRoomRepository.hostStopScreenShare(roomData.roomId, [userId]);
  }

  @override
  String? getScreenSharingUserId() => roomData.roomInfo?.screenSharingUserId;

  @override
  bool isOtherSharing() {
     final uid = getScreenSharingUserId();
     return uid != null && uid != roomData.myUserId;
  }

  @override
  bool isSharingScreen() {
    final uid = getScreenSharingUserId();
    return uid != null && uid == roomData.myUserId;
  }

  @override
  Future<NEResult<void>> startScreenShare([NERtcScreenConfig? config, String? broadcastAppGroup]) {
    apiLogger.i('startScreenShare: profile=${config?.videoProfile} fps=${config?.frameRate} appGroup=$broadcastAppGroup');
    if (getScreenSharingUserId() != null) {
      return Future.value(NEResult(code: RoomErrorCode.failed));
    } else {
      Timer.run(() {
        realStartScreenShare(config, broadcastAppGroup);
      });
      startScreenShareCompleter ??= Completer();
      return startScreenShareCompleter!.future;
    }
  }

  void realStartScreenShare([NERtcScreenConfig? config, String? broadcastAppGroup]) async {
    if (isIOSPlatform()) {
      if (broadcastAppGroup == null) {
        notifyStartScreenShareResult(NEResult(code: RoomErrorCode.paramsError));
        return;
      }
      iosScreenSharePluginSetup(broadcastAppGroup);
      await iosScreenSharePlugin!.launchBroadcastUI();
      await iosScreenSharePlugin!.broadcastStartedFuture;
    }

    final screenShareConfig = config ?? NERtcScreenConfig();
    var error = await NERoomMediaService.instance.engine.startScreenCapture(screenShareConfig);
    if (error != 0) {
      commonLogger.i('engine startScreenCapture error: $error');
      stopLocalScreenShare();
      notifyStartScreenShareResult(NEResult(code: error));
    } else {
      commonLogger.i('engine startScreenCapture succeed');
      final result = await InRoomRepository.startScreenShare(roomData.roomId);
      if (!result.isSuccess()) {
        stopLocalScreenShare();
      }
      notifyStartScreenShareResult(result);
    }
  }

  void notifyStartScreenShareResult(NEResult result) {
    if (startScreenShareCompleter != null && !startScreenShareCompleter!.isCompleted) {
      startScreenShareCompleter!.complete(result);
      startScreenShareCompleter = null;
    }
  }

  @override
  Future<NEResult<void>> stopScreenShare() async {
    apiLogger.i('stopScreenShare');
    if (!isSharingScreen()) {
      return Future.value(NEResult(code: RoomErrorCode.failed));
    } else {
      final result = await InRoomRepository.stopScreenShare(roomData.roomId);
      if (result.isSuccess()) {
        stopLocalScreenShare();
      }
      return result;
    }
  }

  void stopLocalScreenShare() {
    if (isIOSPlatform()) {
      iosScreenSharePlugin?.stopBroadcast();
      iosScreenSharePlugin?.clearBroadcastCallback();
      iosScreenSharePlugin = null;
    }
    NERoomMediaService.instance.engine.stopScreenCapture();
  }

  bool isIOSPlatform() {
    return defaultTargetPlatform == TargetPlatform.iOS;
  }

  void iosScreenSharePluginSetup(String appGroup) {
    if (isIOSPlatform() && iosScreenSharePlugin == null) {
      iosScreenSharePlugin =
          NERtcScreenShare(appGroup)
            ..setBroadcastCallback(this);
    }
  }

  void release() {
    stopLocalScreenShare();
  }

  @override
  void onScreenShareBroadcastFinished() {
    commonLogger.i('onIosScreenShareBroadcastFinished');
    if (isSharingScreen()) {
      stopScreenShare();
    }
  }

  @override
  void onScreenShareBroadcastStarted() {
    commonLogger.i('onIosScreenShareBroadcastStarted');
  }
}
