// Copyright (c) 2014-2021 NetEase, Inc.
// All right reserved.
part of room_kit;

/// 控制房间内自己、其他人的视频状态
class _NEInRoomVideoControllerImpl extends NEInRoomVideoController with _RoomAloggerMixin {

  final _InRoomData roomData;

  _NEInRoomVideoControllerImpl(this.roomData);

  Future<void> setUp() async {
    await syncLocalVideoStatus();
  }

  @override
  Future<NEResult> syncLocalVideoStatus() async {
    commonLogger.i('syncLocalVideoStatus: ${roomData.self.videoStatus}');
    var status = await Permission.camera.status;
    var permissionGranted = status == PermissionStatus.granted;
    if (!permissionGranted){
      commonLogger.i('syncLocalVideoStatus: no permission');
      return  NEResult(code:RoomErrorCode.failed);
    }
    var result = await NERoomMediaService.instance.engine
        .enableLocalVideo(roomData.self.videoStatus == NERoomVideoStatus.on);
    commonLogger.i('syncLocalVideoStatus: result $result');
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  bool canSwitchCamera() {
    // todo check
    return true;
  }

  @override
  bool canUnmuteMyVideo() =>
      roomData.self.videoStatus != NERoomVideoStatus.mutedByHost;

  @override
  String? getPinnedUserId() =>
      roomData.roomInfo?.pinnedUserId;

  // @override
  // String getSelectedCameraId() {
  //   // TODO: implement getSelectedCameraId
  //   throw UnimplementedError();
  // }
  //
  // @override
  // bool isBackCamera(String cameraId) {
  //   // TODO: implement isBackCamera
  //   throw UnimplementedError();
  // }
  //
  // @override
  // bool isFrontCamera(String cameraId) {
  //   // TODO: implement isFrontCamera
  //   throw UnimplementedError();
  // }

  @override
  bool isMyVideoMuted() =>
      roomData.self.videoStatus != NERoomVideoStatus.on;

  @override
  bool isUserPinned(String userId) =>
      roomData.roomInfo?.pinnedUserId == userId;

  @override
  Future<NEResult<void>> muteMyVideo(bool mute) {
    apiLogger.i('muteMyVideo: mute=$mute');
    return mute
        ? InRoomRepository.muteSelfVideo(roomData.roomId)
        : InRoomRepository.unMuteSelfVideo(roomData.roomId);
  }

  @override
  Future<NEResult<void>> pinVideo(String userId, bool pin) {
    apiLogger.i('pinVideo: userId=$userId pin=$pin');
    return pin
        ? InRoomRepository.hostSetFocusVideo(roomData.roomId, [userId])
        : InRoomRepository.hostCancelFocusVideo(roomData.roomId, [userId]);
  }

  @override
  Future<NEResult<void>> subscribeRemoteVideoStream(String userId, int streamType) async {
    final result = await toggleSubscribeRemoteVideoStream(userId, streamType, true);
    apiLogger.i('subscribeRemoteVideoStream: $userId $streamType $result');
    return NEResult(code: result);
  }

  @override
  Future<NEResult<void>> unsubscribeRemoteVideoStream(String userId) async {
    final result = await toggleSubscribeRemoteVideoStream(userId, NERtcRemoteVideoStreamType.low, false);
    apiLogger.i('unsubscribeRemoteVideoStream: $userId $result');
    return NEResult(code: result);
  }

  Future<int> toggleSubscribeRemoteVideoStream(String userId, int streamType, bool subscribe) async {
    if (userId == roomData.self.userId) {
      return RoomErrorCode.paramsError;
    }
    final rtcId = roomData.userId2User[userId]?.rtcId;
    var result = RoomErrorCode.paramsError;
    if (rtcId != null) {
      result = await NERoomMediaService.instance.engine.subscribeRemoteVideo(rtcId, streamType, subscribe);
    }
    return result == 0 ? RoomErrorCode.success : result;
  }

  @override
  Future<NEResult<void>> subscribeRemoteVideoSubStream(String userId) async {
    final result = await toggleSubscribeRemoteVideoSubStream(userId, true);
    apiLogger.i('subscribeRemoteVideoSubStream: $userId $result');
    return NEResult(code: result);
  }

  @override
  Future<NEResult<void>> unsubscribeRemoteVideoSubStream(String userId) async {
    final result = await toggleSubscribeRemoteVideoSubStream(userId, false);
    apiLogger.i('unsubscribeRemoteVideoSubStream: $userId $result');
    return NEResult(code: result);
  }

  Future<int> toggleSubscribeRemoteVideoSubStream(String userId, bool subscribe) async {
    if (userId == roomData.self.userId) {
      return RoomErrorCode.paramsError;
    }
    final rtcId = roomData.userId2User[userId]?.rtcId;
    var result = RoomErrorCode.paramsError;
    if (rtcId != null) {
      result = await NERoomMediaService.instance.engine.subscribeRemoteSubStreamVideo(rtcId, subscribe);
    }
    return result == 0 ? RoomErrorCode.success : result;
  }

  @override
  Future<bool> switchCamera() async {
    return await NERoomMediaService.instance.engine.deviceManager.switchCamera() == 0;
  }

  @override
  Future<NEResult<void>> askParticipantStartVideo(String userId) {
    apiLogger.i('askParticipantStartVideo: $userId');
    if (roomData.self.userId == userId) {
      return InRoomRepository.unMuteSelfVideo(roomData.roomId);
    } else {
     return InRoomRepository.hostUnMuteVideo(roomData.roomId, [userId]);
    }
  }

  @override
  Future<NEResult<void>> stopParticipantVideo(String userId) {
    apiLogger.i('stopParticipantVideo: $userId');
    if (roomData.self.userId == userId) {
      return InRoomRepository.muteSelfVideo(roomData.roomId);
    } else {
      return InRoomRepository.hostMuteVideo(roomData.roomId, [userId]);
    }
  }

  @override
  Future<NEResult<void>> attachRendererToUserVideoStream(NERtcVideoRenderer renderer, String userId) async {
    var result = RoomErrorCode.paramsError;
    if (roomData.userId2User.containsKey(userId)) {
       if (roomData.self.userId == userId) {
         result = await renderer.attachToLocalVideo();
       } else {
         result = await renderer.attachToRemoteVideo(roomData.userId2User[userId]!.rtcId);
       }
    }
    apiLogger.i('attachRendererToUserVideoStream $userId $result');
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }

  @override
  Future<NEResult<void>> attachRendererToUserVideoSubStream(NERtcVideoRenderer renderer, String userId) async {
    var result = RoomErrorCode.paramsError;
    if (roomData.userId2User.containsKey(userId)) {
      if (roomData.self.userId == userId) {
        result = await renderer.attachToLocalSubStreamVideo();
      } else {
        result = await renderer.attachToRemoteSubStreamVideo(roomData.userId2User[userId]!.rtcId);
      }
    }
    apiLogger.i('attachRendererToUserVideoSubStream $userId $result');
    return NEResult(code: result == 0 ? RoomErrorCode.success : result);
  }


}
