// Copyright (c);; 2014-2020 NetEase, Inc.
// All right reserved.
part of room_kit;

/// 在房间里获取白板控制器实例.
class _NEInRoomWhiteboardControllerImpl extends NEInRoomWhiteboardController
    with _RoomAloggerMixin {

  final _InRoomData roomData;
  final List<NEInRoomServiceListener> listeners;

  _NEInRoomWhiteboardControllerImpl(this.roomData, this.listeners);

  @override
  bool hasInteractPrivilege() =>
      roomData.self.whiteBoardInteractStatus ==
          NERoomWhiteBoardInteractionStatus.open;

  @override
  bool hasInteractPrivilegeWithUserId(String userId) {
    var user = roomData.userId2User[userId];
    return user == null
        ? false
        : user.whiteBoardInteractStatus ==
        NERoomWhiteBoardInteractionStatus.open;
  }

  @override
  Future<NEResult<void>> setInteractPrivilege(String userId,
      bool enable) async {
    var user = roomData.userId2User[userId];
    if (user == null) return NEResult(code: RoomErrorCode.failed);

    if (enable) {
      return await InRoomRepository.awardedMemberWhiteboardInteraction(
          roomData.roomInfo!.roomId, [user.rtcId]);
    }
    return await InRoomRepository.undoMemberWhiteboardInteraction(
        roomData.roomInfo!.roomId, [user.rtcId]);
  }

  @override
  Future<NEResult<void>> showWhiteboardTools(bool show) async {
    /// 没有权限，不允许开白板
    if (!isLogin() ||
        roomData.self.whiteBoardInteractStatus !=
            NERoomWhiteBoardInteractionStatus.open) {
      return NEResult(code: RoomErrorCode.failed);
    }
    enableDraw(show);
    return NEResult.success();
  }

  @override
  JavascriptChannel initWhiteboardJavascriptChannel(
    Future<WebViewController> javascriptController,
    NEWhiteboardErrorCallback? onWhiteboardErrorCallback,
  ) {
    return NEWhiteBoard.instance.initWhiteboardJavascriptChannel(
      javascriptController,
      onWhiteboardPageLoaded:(){
        final info = roomData.roomInfo;
        if (info == null) {
          commonLogger.e('roomInfo is null');
          if (onWhiteboardErrorCallback != null) {
            onWhiteboardErrorCallback.call(code: RoomErrorCode.failed);
          }
          return;
        }

        var whiteBoardUserInfo = RoomWBLoginInfo(
          appKey: roomData.joinRoomInfo.wbKey!,
          debug: false,
          nickname: UserProfile.displayName,
          record: roomData.joinRoomInfo.attendeeRecordOn,
          platform: _isIOSPlatform() ? iOS : android,
          channelName: info.uniqueId.toString(),
          token: roomData.joinRoomInfo.wbToken!,
          account: roomData.joinRoomInfo.wbAccid!,
        );
        commonLogger.i('whiteBoardUserInfo ${whiteBoardUserInfo.toString()}');
        NEWhiteBoard.instance.login(whiteBoardUserInfo);
      },
      onWhiteboardJoinWBSucceed: () => webJoinWBSucceed(),
      onWhiteboardErrorCallback: ({required code, msg}) {
        if (onWhiteboardErrorCallback != null) {
          onWhiteboardErrorCallback.call(code: code, msg: msg);
        }
      },
      onWhiteboardInterceptLog: (log) {
        commonLogger.i('flutter NativeFunction javascript dataFromJs： $log');
      },
    );
  }

  @override
  bool isOtherSharing() {
    final userId = getWhiteboardSharingUserId();
    return userId != null && userId != roomData.self.userId;
  }

  @override
  bool isSharingWhiteboard() {
    return roomData.self.isWhiteBoardSharing;
  }

  @override
  bool isWhiteboardSharing(String userId) =>
      userId == roomData.roomInfo?.whiteboardSharingUserId;

  @override
  String? getWhiteboardSharingUserId() {
    return roomData.roomInfo?.whiteboardSharingUserId;
  }

  @override
  void release() async {
    await NEWhiteBoard.instance.release();
  }

  void webJoinWBSucceed() {

    var whiteResult = NEWhiteBoard.instance.setContainerOptions();
    commonLogger.i(' webJoinWBSucceed jsSetContainerOptions$whiteResult');

    var isWhiteBoardOwner =
        roomData.roomInfo!.whiteboardSharingUserId == roomData.self.userId;
    if (isWhiteBoardOwner) {
      roomData.self.whiteBoardInteractStatus =
          NERoomWhiteBoardInteractionStatus.open;
    }
    var enable = roomData.self.whiteBoardInteractStatus ==
        NERoomWhiteBoardInteractionStatus.open;
    enableDraw(enable);
  }

  ///默认不显示白板工具栏
  void enableDraw(bool enable) async {
    final result = await NEWhiteBoard.instance.enableDraw(enable);
    commonLogger.i('enableDraw enable: $enable, result: $result');
  }

  bool isLogin() => NEWhiteBoard.instance.isWhiteboardLogin();

  @override
  Future<NEResult<void>> startWhiteboardShare() {
    apiLogger.i('startWhiteboardShare');
    var result = InRoomRepository.startWhiteBoardShare(roomData.roomId);
    return result;
  }

  @override
  Future<NEResult<void>> stopWhiteboardShare() async {
    apiLogger.i('stopWhiteboardShare');
    return await InRoomRepository.stopWhiteBoardShare(roomData.roomId);
  }

  @override
  Future<NEResult<void>> stopParticipantWhiteboardShare(String userId) async {
    apiLogger.i('stopParticipantScreenShare userId $userId');
    return await InRoomRepository.hostStopWhiteBoardShare(
        roomData.roomId, [userId]);
  }

  @override
  String? getWhiteboardUrl() => NEWhiteBoard.instance.getWhiteboardUrl();
}

bool _isIOSPlatform() => defaultTargetPlatform == TargetPlatform.iOS;