// Copyright (c); 2014-2020 NetEase, Inc.
// All right reserved.
part of room_kit;

/// 在房间里获取聊天室控制器实例.
abstract class NEInRoomChatController {

  /// 查询当前会议聊天室是否可用
  bool isChatroomEnabled();

  ///查询是否在当前房间中开启聊天
  bool hasEnterChatRoom();

  ///加入聊天室
  Future<NEResult<void>> enterChatRoom();

  ///退出聊天室
  Future<NEResult<void>> exitChatRoom();

  /// 更新聊天室用户角色信息
  Future<NEResult<void>> updateChatRoomUserRole(
      NEChatRoomUserRole chatRoomUserInfo,
      {Map<String, dynamic>? extension});

  Future<NEResult<NEChatRoomMessage>> sendChatRoomMessage(NEChatRoomMessage message);
}

class NEChatRoomUserRole {
  String? nick;
  String? avatar;

  NEChatRoomUserRole({required this.nick, this.avatar});
}

class NEChatRoomUserInfo extends NEChatRoomUserRole {
  bool isSelf = false;

  NEChatRoomUserInfo(
      {String? nick, String? avatar, required this.isSelf})
      : super(
          nick: nick,
          avatar: avatar,
        );
}

abstract class NEChatRoomMessage {
  ///内部设置私有属性
  NEChatRoomUserInfo? _chatRoomUserInfo;

  /// 获取聊天室消息扩展属性
  ///
  /// @return 聊天室消息扩展属性
  Map<String, dynamic>? senderExtension;

  ///聊天消息内容
  String? content;

  int? time;

  /// [MsgType]
  int? _msgType;

  /// [MsgStatus]
  int? _msgStatus;

  NEChatRoomUserInfo get chatRoomUserInfo =>
      _chatRoomUserInfo ?? NEChatRoomUserInfo(isSelf: true);

  int get msgType => _msgType ?? MsgType.undef;

  int get msgStatus => _msgStatus ?? MsgStatus.draft;

  NEChatRoomMessage({
    this.senderExtension,
    this.content,
    this.time,
  });

  @override
  String toString() {
    return 'NEChatRoomMessage{ senderExtension: $senderExtension, content: $content, time: $time ,msgType $msgType}';
  }
}

class NEChatRoomTextMessage extends NEChatRoomMessage {
  NEChatRoomTextMessage({
    Map<String, dynamic>? senderExtension,
    String? content,
    int? time,
  }) : super(
          senderExtension: senderExtension,
          content: content,
          time: time,
        );

  @override
  int get msgType => MsgType.text;
}

class NEChatRoomImageMessage extends NEChatRoomMessage {
  final NEImageAttachment _attachment;

  NEImageAttachment get attachment => _attachment;

  NEChatRoomImageMessage({
    Map<String, dynamic>? senderExtension,
    String? content,
    int? time,
    int? width,
    int? height,
    required String path,
    required String displayName,
  })   : _attachment = NEImageAttachment(
          width: width,
          height: height,
          url: path,
          displayName: displayName,
        ),
        super(
          senderExtension: senderExtension,
          content: content,
          time: time,
        );

  @override
  int get msgType => MsgType.image;

  @override
  String toString() {
    return 'NEChatRoomImageMessage{_attachment: $_attachment}';
  }
}

class NEChatRoomFileMessage extends NEChatRoomMessage {
  ///消息发送成功后的附件
  final NEFileAttachment _attachment;

  NEFileAttachment get attachment => _attachment;

  @override
  int get msgType => MsgType.file;

  NEChatRoomFileMessage({
    Map<String, dynamic>? senderExtension,
    String? content,
    int? time,
    required String path,
    required String displayName,
  })   : _attachment = NEFileAttachment(
          url: path,
          displayName: displayName,
        ),
        super(
          senderExtension: senderExtension,
          content: content,
          time: time,
        );

  @override
  String toString() {
    return 'NEChatRoomFileMessage{_attachment: $_attachment}';
  }
}

/// source form [FileAttachment]
class NEFileAttachment {
  /// 文件大小
  int? size;

  /// 文件内容md5
  String? md5;

  /// 文件下载地址
  String url;

  /// 文件显示名
  String displayName;

  /// 文件后缀名
  String? extension;

  /// 过期时间（仅对文件类型有效）
  int? expire;

  NEFileAttachment({
    this.size,
    this.md5,
    required this.url,
    required this.displayName,
    this.extension,
    this.expire,
  });

  @override
  String toString() {
    return 'NEFileAttachment{size: $size, md5: $md5, url: $url, displayName: $displayName, extension: $extension, expire: $expire}';
  }
}

/// source form [ImageAttachment]
class NEImageAttachment extends NEFileAttachment {
  /// 图片宽
  int? width;

  /// 图片高
  int? height;

  NEImageAttachment({
    this.width,
    this.height,
    int? size,
    String? md5,
    required String url,
    required String displayName,
    String? extension,
    int? expire,
  }) : super(
          size: size,
          md5: md5,
          url: url,
          displayName: displayName,
          extension: extension,
          expire: expire,
        );

  @override
  String toString() {
    return 'NEImageAttachment{width: $width, height: $height}';
  }
}
