// Copyright (c) 2014-2020 NetEase, Inc.
// All right reserved.
part of room_kit;

class _NEAccountServiceImpl extends NEAuthService
    with _RoomAloggerMixin, EventTrackMixin, _RoomAloggerMixin {

  static final _NEAccountServiceImpl _instance = _NEAccountServiceImpl._();

  factory _NEAccountServiceImpl() => _instance;

  _NEAccountServiceImpl._() {
    /// 账号token过期监听，如修改密码
    InRoomRepository.onServerControlReceived.listen((MeetingAction action) {
      if (action is AuthInfoExpiredAction) {
        commonLogger.i('auth info expired from server notify');
        _notifyAuthInfoExpired();
      }
    });
    /// 透传协议因为网络原因失败，监听其他协议的失败
    RoomErrorRepository.authInfoExpiredError.listen((event) {
      commonLogger.i('auth info expired from api error');
      _notifyAuthInfoExpired();
    });
    AuthRepository.onlineStatusChanged.listen((status) async {
      commonLogger.i('online status changed: $status');
      if (status == AuthState.kickOut) {
        _notifyKickOut();
      }
    });
  }

  final Set<NERoomAuthListener> _authListenerSet = <NERoomAuthListener>{};

  @override
  NEAccountInfo? getAccountInfo() {
    final meetingId = UserProfile.meetingId;
    final shortId = UserProfile.shortId;
    final accountId = UserProfile.accountId;
    final nick = UserProfile.nickname;
    return meetingId != null && accountId != null && nick != null
        ? NEAccountInfo(accountId, nick, meetingId, shortId)
        : null;
  }
  @override
  Future<NEResult<void>> loginWithToken(String accountId, String token) async {
    apiLogger.i('loginWithToken accountId $accountId');
    trackPeriodicEvent(TrackEventName.loginSdk,
        extra: {'value': _TrackExtra.token});
    if (isCurrentRoomConnected()) return onRoomIsUnderGoing();
    //登录前重置isAnonymous，因为匿名入会出来，isAnonymous有可能为true
    UserProfile.isAnonymous = false;
    Future<NEResult<void>> onAccept() async {
      return trackLoginResultEvent(
          await AuthRepository.loginWithToken(accountId, token));
    }

    return InvocationManager().asyncWithResource(
      [Resource.account],
      onAccept,
      onDeny,
    );
  }

  @override
  Future<NEResult<void>> loginWithAccount(String username, String password) {
    apiLogger.i('loginWithNEMeeting');
    if (isCurrentRoomConnected()) return onRoomIsUnderGoing();
    //登录前重置isAnonymous，因为匿名入会出来，isAnonymous有可能为true
    UserProfile.isAnonymous = false;
    Future<NEResult<void>> onAccept() async {
      trackPeriodicEvent(TrackEventName.loginSdk,
          extra: {'value': _TrackExtra.password});
      return trackLoginResultEvent(
          await AuthRepository.loginWithNEMeeting(username, password));
    }

    return InvocationManager().asyncWithResource(
      [Resource.account],
      onAccept,
      onDeny,
    );
  }

  @override
  Future<NEResult<void>> loginWithSSOToken(String ssoToken) {
    apiLogger.i('loginWithSSOToken');
    if (isCurrentRoomConnected()) return onRoomIsUnderGoing();
    //登录前重置isAnonymous，因为匿名入会出来，isAnonymous有可能为true
    UserProfile.isAnonymous = false;

    Future<NEResult<void>> onAccept() async {
      trackPeriodicEvent(TrackEventName.loginSdk,
          extra: {'value': _TrackExtra.sso});
      return trackLoginResultEvent(
          await AuthRepository.loginWithSSOToken(ssoToken));
    }

    return InvocationManager().asyncWithResource(
      [Resource.account],
      onAccept,
      onDeny,
    );
  }

  @override
  Future<NEResult<void>> tryAutoLogin() {
    apiLogger.i('tryAutoLogin');
    if (isCurrentRoomConnected()) return onRoomIsUnderGoing();
    //登录前重置isAnonymous，因为匿名入会出来，isAnonymous有可能为true
    UserProfile.isAnonymous = false;

    Future<NEResult<void>> onAccept() async {
      trackPeriodicEvent(TrackEventName.loginSdk,
          extra: {'value': _TrackExtra.auto});
      return trackLoginResultEvent(await AuthRepository.tryAutoLogin());
    }

    return InvocationManager()
        .asyncWithResource([Resource.account], onAccept, onDeny);
  }

  @override
  void addAuthListener(NERoomAuthListener listener) {
    apiLogger.i('addAuthListener');
    _authListenerSet.add(listener);
  }

  @override
  void removeAuthListener(NERoomAuthListener listener) {
    apiLogger.i('removeAuthListener');
    _authListenerSet.remove(listener);
  }

  @override
  Future<NEResult<void>> logout() {
    apiLogger.i('logout');
    if (isCurrentRoomConnected()) return onRoomIsUnderGoing();
    Future<NEResult<void>> onAccept() {
      trackPeriodicEvent(TrackEventName.loginOutSdk);
      SettingsRepository.clear();
      if (AuthRepository.isAuthorized()) {
        AuthRepository.logout();
      }
      return Future.value(NEResult(code: RoomErrorCode.success));
    }

    return InvocationManager().syncWithResource(
      [Resource.account],
      onAccept,
      onDeny,
    );
  }

  void _notifyKickOut() {
    _authListenerSet.toList().forEach((listener) {
      listener.onKickOut();
    });
  }

  void _notifyAuthInfoExpired() {
    AuthRepository.logout();
    _authListenerSet
        .toList(growable: false)
        .forEach((listener) => listener.onAuthInfoExpired());
  }

  bool isCurrentRoomConnected() {
    final roomService = NERoomKit.instance.getRoomService();
    if (roomService is _NERoomServiceImpl) {
      return roomService.isCurrentRoomConnected();
    }
    return false;
  }

  NEResult<void> trackLoginResultEvent(NEResult<void> result) {
    if (result.code == RoomErrorCode.success) {
      trackPeriodicEvent(TrackEventName.loginSdkSuccess);
    } else {
      trackPeriodicEvent(TrackEventName.loginSdkFailed,
          extra: {'value': result.code});
    }
    return result;
  }
}
