// Copyright (c) 2014-2021 NetEase, Inc.
// All right reserved.

part of room_kit;

class _NEInRoomServiceImpl extends NEInRoomService
    with
        NERtcChannelEventCallback,
        NERtcStatsEventCallback,
        NERtcDeviceEventCallback,
        EventTrackMixin,
        _RoomAloggerMixin {
  late final _InRoomData roomData;

  late final NEInRoomCloudRecordController cloudRecordController;
  late final _NEInRoomAudioControllerImpl audioController;
  late final _NEInRoomChatControllerImpl chatController;
  late final NEInRoomLiveStreamController liveStreamController;
  late final _NEInRoomScreenShareControllerImpl screenShareController;
  late final _NEInRoomVideoControllerImpl videoController;
  late final _NEInRoomWhiteboardControllerImpl whiteboardController;
  late final _NEInRoomBeautyControllerImpl beautyController;
  late final NEInRoomUserController userController;

  /// 当前是否为匿名入会
  final bool anonymous;

  /// 当前入会信息
  final JoinRoomInfo joinRoomInfo;

  static const initial = -1;
  static const connecting = 0;
  static const connected = 1;
  static const disconnecting = 2;
  static const connectFail = 3;
  static const disconnected = 4;

  var state = initial;
  bool get isActive => state <= connected;
  bool isStateAtLeast(int target) {
    return state >= target;
  }

  final listeners = <NEInRoomServiceListener>[];

  final streamSubscriptions = <StreamSubscription>[];

  final rtcStatsEventListeners = <NERtcStatsEventListener>{};
  final rtcDeviceEventListeners = <NERtcDeviceEventCallback>{};

  final lifecycleOwner = NERoomLifeCycleObserverOwner();

  /// 控制只有在 connected 之后，所有的变更通知才往外发
  final _RoomInfoChangeNotifyDecider changeNotifyDecider = _RoomInfoChangeNotifyDecider();

  /// 抑制 音频 状态变更通知
  /// 主持人全体静音时，会对用户列表中的所有成员修改音频状态为muteByHost，这个时候不会对外发送通知
  /// 因为可能房间内包含较多的用户数，此时每个用户都需要触发一次该回调，性能会有影响
  bool suppressAudioStatusChangeNotify = false;

  Timer? checkAVStatusConsistencyTask;

  final Completer _completer = Completer();
  Future get finished => _completer.future;
  void didFinish() {
    if (!_completer.isCompleted) {
      changeNotifyDecider.dispose();
      rtcStatsEventListeners.clear();
      rtcDeviceEventListeners.clear();
      checkAVStatusConsistencyTask?.cancel();
      listeners.clear();
      streamSubscriptions.forEach((element) {
        element.cancel();
      });
      if (anonymous) {
        AuthRepository.logout();
      }
      lifecycleOwner
        ..notifyRoomOnIdle()
        ..clear();
      _completer.complete();    
    }
  }

  /// 入会需要传入的option
  final NERoomOptions options;

  _NEInRoomServiceImpl(this.anonymous, this.joinRoomInfo, this.options) {
    UserProfile.lastMeetingId = joinRoomInfo.meetingId;
    UserProfile.lastChannelId = int.parse(joinRoomInfo.avRoomCid);
    roomData = _InRoomData(joinRoomInfo, options);

    cloudRecordController = _NEInRoomCloudRecordControllerImpl(roomData);
    audioController = _NEInRoomAudioControllerImpl(roomData, this);
    videoController = _NEInRoomVideoControllerImpl(roomData);
    liveStreamController = _NEInRoomLiveStreamControllerImpl(roomData);
    screenShareController = _NEInRoomScreenShareControllerImpl(roomData);
    whiteboardController = _NEInRoomWhiteboardControllerImpl(roomData,listeners);
    beautyController = _NEInRoomBeautyControllerImpl(roomData);
    chatController = _NEInRoomChatControllerImpl(roomData,listeners);
    userController = _NEInRoomUserControllerImpl(roomData);
  }

  void registerEventListener() {
    streamSubscriptions.add(AuthRepository.imOnlineStatusChanged.listen((dynamic code) {
      if (code == StatusCode.logined) {
        fetchRoomData();
      }
    }));
    streamSubscriptions.add(RoomErrorRepository.authInfoExpiredError.listen((event) {
      notifyRoomDisconnected(NERoomDisconnectReason.authInfoExpired);
    }));
    streamSubscriptions.add(AuthRepository.onlineStatusChanged.listen((status) {
      if (status == AuthState.kickOut) {
        notifyRoomDisconnected(NERoomDisconnectReason.loginOnOtherDevice);
      }
    }));
  }

  void notifyRoomConnectFail(int reason) {
    if (isStateAtLeast(connectFail)) return;
    state = connectFail;
    commonLogger.i('onRoomConnectFail: $reason');
    lifecycleOwner.notifyRoomConnectFail(reason);
    Timer.run(() {
      release();
    });
  }

  void notifyRoomDisconnecting() {
    if (isStateAtLeast(disconnecting)) return;
    state = disconnecting;
    commonLogger.i('onRoomDisconnecting');
    lifecycleOwner.notifyRoomOnDisConnecting();
  }

  void notifyRoomDisconnected(int reason) {
    if (isStateAtLeast(disconnected)) return;
    state = disconnected;
    commonLogger.i('onRoomDisconnected: $reason');
    lifecycleOwner.notifyRoomOnDisConnected(reason);
    Timer.run(() {
      release();    
    });
  }

  @override
  void startConnect() {
    if (isStateAtLeast(connecting)) {
      return;
    }
    apiLogger.i('startConnect');
    registerEventListener();
    registerServerControl(joinRoomInfo.avRoomCid);
    prepareRoomData();
  }

  @override
  void addListener(NEInRoomServiceListener listener) => listeners.add(listener);

  @override
  void removeListener(NEInRoomServiceListener listener) =>
      listeners.remove(listeners);

  @override
  void addRoomLifeCycleObserver(NERoomLifeCycleObserver observer) {
    apiLogger.i('addRoomLifeCycleObserver');
    lifecycleOwner.addObserver(observer);
  }

  @override
  void removeRoomLifeCycleObserver(NERoomLifeCycleObserver observer) {
    apiLogger.i('removeRoomLifeCycleObserver');
    lifecycleOwner.removeObserver(observer);
  }

  @override
  String getCurrentRoomId() => joinRoomInfo.meetingId;

  @override
  int getUserCount() =>
      roomData.userId2User.values
      .where((user) => RoleType.isHide(user.roleType))
      .length;

  @override
  NEInRoomAudioController getInRoomAudioController() => audioController;

  @override
  NEInRoomBeautyController getInRoomBeautyController() => beautyController;

  @override
  NEInRoomChatController getInRoomChatController() => chatController;

  @override
  NEInRoomCloudRecordController getInRoomCloudRecordController() =>
      cloudRecordController;

  @override
  NEInRoomLiveStreamController getInRoomLiveStreamController() =>
      liveStreamController;

  @override
  NEInRoomScreenShareController getInRoomScreenShareController() =>
      screenShareController;

  @override
  NEInRoomVideoController getInRoomVideoController() => videoController;

  @override
  NEInRoomWhiteboardController getInRoomWhiteboardController() =>
      whiteboardController;

  @override
  NEInRoomUserController getInRoomUserController() => userController;

  @override
  String getMyUserId() => roomData.myUserId;

  int? getRtcId(String userId) => roomData.userId2User[userId]?.rtcId;

  @override
  NEInRoomUserInfo? getMyUserInfo() => getUserInfoById(getMyUserId());

  @override
  NEInRoomUserInfo? getUserInfoById(String userId) =>
      roomData.userId2User[userId];

  @override
  bool isHostUser(String userId) => userId == roomData.roomInfo?.hostUserId;

  @override
  bool isMySelfHost() => getMyUserId() == roomData.roomInfo?.hostUserId;

  @override
  String? getHostUserId() => roomData.roomInfo?.hostUserId;

  @override
  bool isMyself(String userId) => userId == getMyUserId();

  @override
  bool isParticipantsUnmuteSelfAllowed() =>
      roomData.roomInfo?.audioAllMute != true ||
      roomData.roomInfo?.screenSharingUserId == roomData.self.userId ||
      roomData.roomInfo?.settings?.isAllowSelfAudioOn != false;

  @override
  bool isRoomLocked() => roomData.roomInfo?.locked == true;

  @override
  Future<NEResult<void>> leaveCurrentRoom(bool closeIfHost) async {
    NEResult<void>? result;
    if (isActive) {
      apiLogger.i('leaveCurrentRoom: closeIfHost=$closeIfHost');
      notifyRoomDisconnecting();
      var reason = NERoomDisconnectReason.leaveBySelf;
      if (isMySelfHost() && closeIfHost) {
        reason = NERoomDisconnectReason.closeBySelf;
        result = await RoomRepository.closeRoom(getCurrentRoomId());
        commonLogger.i('leaveCurrentRoom: closeRoomResult=${result.isSuccess()}');
      }
      notifyRoomDisconnected(reason);
    }
    return result ?? NEResult.success();
  }

  @override
  Future<NEResult<void>> lockRoom(bool isLock) {
    apiLogger.i('lockRoom: $isLock');
    return isLock
        ? InRoomRepository.hostLockMeeting(getCurrentRoomId())
        : InRoomRepository.hostUnLockMeeting(getCurrentRoomId());
  }

  @override
  NERoomInfo? getCurrentRoomInfo() => roomData.roomInfo;

  @override
  void addRtcStatsEventListener(NERtcStatsEventListener listener)  {
    apiLogger.i('addRtcStatsEventListener');
    if(rtcStatsEventListeners.isEmpty){
       NERoomMediaService.instance.engine.setStatsEventCallback(this);
    }
    rtcStatsEventListeners.add(listener);
  }

  @override
  void removeRtcStatsEventListener(NERtcStatsEventListener listener)  {
    apiLogger.i('removeRtcStatsEventListener ->');
    rtcStatsEventListeners.remove(listener);
    if(rtcStatsEventListeners.isEmpty){
       NERoomMediaService.instance.engine.clearStatsEventCallback();
    }
  }

  @override
  void addRtcDeviceEventListener(NERtcDeviceEventCallback listener)  {
    apiLogger.i('addRtcDeviceEventListener');
    if(rtcDeviceEventListeners.isEmpty){
       NERoomMediaService.instance.engine.deviceManager.setEventCallback(this);
    }
    rtcDeviceEventListeners.add(listener);
  }

  @override
  void removeRtcDeviceEventListener(NERtcDeviceEventCallback listener)  {
    apiLogger.i('removeRtcDeviceEventListener');
    rtcDeviceEventListeners.remove(listener);
    if(rtcDeviceEventListeners.isEmpty){
       NERoomMediaService.instance.engine.deviceManager.clearEventCallback();
    }
  }

  @override
  List<NEInRoomUserInfo> getAllUsers() =>
      roomData.userId2User.values
          .where((user) => user.isOnline)
          .toList(growable: false);

  void registerServerControl(String avRoomCid) {
    streamSubscriptions.add(InRoomRepository.onServerControlReceived.listen((MeetingAction action) {
      commonLogger.i('receive room action: ${action.runtimeType}');

      if (!isActive) {
        commonLogger.e('reject room action due to inactive.');
        return;
      }

      if (roomData.roomInfo == null) {
        commonLogger.e('reject room action due to room data not ready.');
        return;
      }

      /// 不为空 且 avRoomCid不一致 return
      if (action.roomCid != null && action.roomCid != avRoomCid) {
        commonLogger.e('reject room action due to different room cid.');
        return;
      }

      switch (action.type) {
        case ActionType.controlAudio:
          handleAudioAction(action as AudioAction);
          break;
        case ActionType.controlVideo:
          handleVideoAction(action as VideoAction);
          break;
        case ActionType.controlFocus:
          final focusAction = action as FocusAction;
          roomData.roomInfo?.pinnedUserId =
              focusAction.isFocus ? focusAction.operateUser : null;
          break;
        case ActionType.controlHost:
          roomData.roomInfo?.hostUserId = (action as HostAction).operateUser;
          break;
        case ActionType.updateNick:
          final nickAction = action as NickAction;
          roomData.userId2User[nickAction.userId]?.displayName = nickAction.nick;
          break;
        case ActionType.screenShare:
          handleScreenShareAction(action as ScreenShareAction);
          break;
        case ActionType.memberChange:
          fetchRoomData(getRtcId((action as MemberChangeAction).userId));
          break;
        case ActionType.meetingLock:
          roomData.roomInfo?.locked = JoinControlType.isLock(
              (action as MeetingLockAction).joinControlType);
          break;
        case ActionType.handsUp:
          handleRaiseHandAction(action as HandsUpAction);
          break;
        case ActionType.live:
          handleLiveAction(action as MeetingLiveAction);
          break;
        case ActionType.startWhiteBoardShare:
        case ActionType.stopWhiteBoardShare:
          handleWhiteBoardShareAction(action as WhiteBoardShareAction);
          break;
        case ActionType.awardedMemberWhiteboardInteraction:
        case ActionType.undoMemberWhiteboardInteraction:
          handleWhiteboardInteractionAction(
              action as WhiteboardInteractionAction);
          break;
        case ActionType.authInfoExpired:
          notifyRoomDisconnected(NERoomDisconnectReason.authInfoExpired);
          break;
      }
    }));
  }

  void handleWhiteBoardShareAction(WhiteBoardShareAction action) {
    commonLogger.i('handleWhiteBoardShare action: $action');
    if (TextUtils.isEmpty(action.sharerAvRoomUid.toString())) {
      return;
    }

    ///计算当前白板状态
    NERoomWhiteBoardShareStatus status;
    if (action.type == ActionType.startWhiteBoardShare) {
      status = NERoomWhiteBoardShareStatus.start;
    } else {
      whiteboardController.release();
      status = action.sharerAvRoomUid != action.operatorAvRoomUid
          ? NERoomWhiteBoardShareStatus.stopByHost
          : NERoomWhiteBoardShareStatus.end;
    }
    roomData.roomInfo!.whiteboardSharingUserInfo = _WhiteboardSharingUserInfo(
      sharerUserId: status != NERoomWhiteBoardShareStatus.start
          ? null
          : roomData.rtcId2User[action.sharerAvRoomUid]!.userId,
      sharerImAccId: status != NERoomWhiteBoardShareStatus.start
          ? null
          : action.sharerImAccid,
      sharerRtcId: status != NERoomWhiteBoardShareStatus.start
          ? null
          : action.sharerAvRoomUid,
      operatorRtcId: status != NERoomWhiteBoardShareStatus.start
          ? null
          : action.operatorAvRoomUid,
      whiteboardShareStatus: status,
    );
  }

  void handleWhiteboardInteractionAction(WhiteboardInteractionAction action) {
    commonLogger.i('handleWhiteboardInteractionAction action: $action');
    if (TextUtils.isEmpty(action.sharerAvRoomUid.toString())) {
      return;
    }
    var enable = action.type == ActionType.awardedMemberWhiteboardInteraction;
    roomData.rtcId2User[action.sharedAvRoomUid]?.whiteBoardInteractStatus = enable
        ? NERoomWhiteBoardInteractionStatus.open
        : NERoomWhiteBoardInteractionStatus.close;
  }

  void handleAudioAction(AudioAction action) async {
    var subType = action.subType;
    if (subType == AudioSubType.muteAllOpenOff ||
        subType == AudioSubType.muteAllOpenOn ||
        subType == AudioSubType.unMuteAll
    ) {
      final willAudioAllMute = subType != AudioSubType.unMuteAll;
      final allowSelfAudioOn = subType != AudioSubType.muteAllOpenOff;
      // 只有第一次操作全体静音，才需要抑制成员音频变更通知
      // 后续连续的全体静音操作，需要放开成员音频变更的通知
      if (roomData.roomInfo?.audioAllMute != willAudioAllMute) {
        suppressAudioStatusChangeNotify = true;
      }
      roomData.userId2User.values
          .where((user) => !isHostUser(user.userId))
          .forEach((user) {
        updateUserAudioStatus(user, action.muteAudio);
        /// 解除全体静音 或者 允许参会者自行解除静音 时，将举手状态放下
        if (!willAudioAllMute || allowSelfAudioOn) {
          user.raiseHandDetail = _InRoomRaiseHandDetail.down();
        }
      });
      suppressAudioStatusChangeNotify = false;
      roomData.roomInfo?.audioAllMute = willAudioAllMute;
      roomData.roomInfo?.settings?.allowSelfAudioOn = allowSelfAudioOn;
    } else if (subType == AudioSubType.muteAudioHandsUpOn
        || action.fromUser == action.operateUser) {
      final user = roomData.userId2User[action.operateUser];
      if (user != null) {
        user.audioStatus = indexToAudioStatus(action.muteAudio);
        user.raiseHandDetail = subType == AudioSubType.muteAudioHandsUpOn
            ? _InRoomRaiseHandDetail(NEHandsUpStatus.agree, -1)
            : _InRoomRaiseHandDetail.down();
      }
    } else {
      updateUserAudioStatus(roomData.userId2User[action.operateUser], action.muteAudio);
    }
  }

  void updateUserAudioStatus(_InRoomUser? user, int statusIndex) {
    if (user != null) {
      final beforeValue = user.audioStatus;
      user.audioStatus = indexToAudioStatus(statusIndex);
      /// 特殊处理：主持人多次请求成员打开音频，需要多次通知出去
      /// 如果前后状态不一致，在设置时会自动通知出去，这里只处理前后状态一致但仍需要通知出去的情况
      if (beforeValue == NERoomAudioStatus.waitingOpen && beforeValue == user.audioStatus) {
        commonLogger.i('onRoomUserAudioStatusChanged: ${user.displayName} ${user.audioStatus}');
        listeners.forEach((element) {
          element.onRoomUserAudioStatusChanged(user.userId, NERoomAudioStatus.waitingOpen);
        });
      }
    }
  }

  void handleVideoAction(VideoAction action) {
    final user = roomData.userId2User[action.operateUser];
    if (user != null) {
      final beforeValue = user.videoStatus;
      user.videoStatus = indexToVideoStatus(action.muteVideo);
      /// 特殊处理：主持人多次请求成员打开视频，需要多次通知出去
      if (beforeValue == NERoomVideoStatus.waitingOpen && beforeValue == user.videoStatus) {
        listeners.forEach((element) {
          element.onRoomUserVideoStatusChanged(user.userId, NERoomVideoStatus.waitingOpen);
        });
      }
    }
  }

  void handleScreenShareAction(ScreenShareAction action) {
    if (action.screenSharing == ScreenShareAction.notShare
        && action.user_av_room_uid != action.oper_user_av_room_uid) {
      commonLogger.i('onRoomUserScreenShareStatusChanged: ${action.userId} ${roomData.getUserName(action.userId)} stopByHost');
      listeners.forEach((element) {
        element.onRoomUserScreenShareStatusChanged(action.userId, NERoomScreenShareStatus.stopByHost);
      });
      if (roomData.myUserId == action.userId) {
        //stop local screen share
        screenShareController.stopLocalScreenShare();
      }
    }
    roomData.roomInfo?.screenSharingUserId =
    action.screenSharing == ScreenShareAction.share
        ? action.userId
        : null;
  }

  void handleLiveAction(MeetingLiveAction action) {
    final liveStreamInfo = roomData.roomInfo?.liveStreamInfo;
    if (liveStreamInfo != null) {
      final beforeState = liveStreamInfo.state;

      liveStreamInfo.delegate.title = action.title;
      liveStreamInfo.delegate.state = action.state;
      liveStreamInfo.delegate.liveChatRoomEnable = action.liveChatRoomEnable;
      liveStreamInfo.delegate.liveAVRoomUids = action.liveAVRoomUids;
      liveStreamInfo.delegate.liveLayout = action.liveLayout;
      liveStreamInfo.password = action.password;
      liveStreamInfo.delegate.liveWebAccessControlLevel = action.liveWebAccessControlLevel;
      roomData.roomInfo?.liveStreamInfo = liveStreamInfo.copy();

      final afterState = liveStreamInfo.state;
      if (afterState != beforeState) {
        commonLogger.i('onRoomLiveStreamStatusChanged: $afterState');
        listeners.forEach((element) {
          element.onRoomLiveStreamStatusChanged(afterState);
        });
      }
    } else {
      commonLogger.w('handleLiveAction failed: current live stream info is null');
    }
  }

  void prepareRoomData() async {
    state = connecting;
    lifecycleOwner.notifyRoomOnConnecting();
    late NEResult<void> result;
    for (var index = 0; index < 3; index++) {
      result = await fetchRoomData();
      if (!isActive) return;
      commonLogger.i('prepareRoomData: index=$index, code=${result.code}, msg=${result.msg}');
      if (result.isSuccess() || result.code == RoomErrorCode.roomNotExist) break;
      /// waiting for NIM reconnect
      await Future.delayed(Duration(seconds: (index + 1) * 3));
    }
    if (result.code == RoomErrorCode.roomNotExist) {
      notifyRoomConnectFail(NERoomConnectFailReason.roomNotExist);
    } else if (!result.isSuccess()) {
      notifyRoomConnectFail(NERoomConnectFailReason.syncDataError);
    } else {
      var result = await NERoomMediaService.instance.initRtcEngine(
        rtcChannelEventCallback: this,
        attendeeRecordOn: roomData.joinRoomInfo.attendeeRecordOn,
        mediaType: MediaType.inRoom,
        ///加入不同企业创建会议，必须是当前服务端返回的nrtcAppKey
        nrtcAppKey: roomData.joinRoomInfo.nrtcAppKey,
      );
      commonLogger.i('initRtcEngine: $result');
      if (!isActive) {
        return;
      } 
      if (result == NERtcErrorCode.ok) {
        trackPeriodicEvent(TrackEventName.joinChannel);
        await NERoomMediaService.instance.setConfig();
        result = await NERoomMediaService.instance.joinChannel(roomData.joinRoomInfo);
        if (result != NERtcErrorCode.ok) {
          notifyRoomConnectFail(NERoomConnectFailReason.joinChannelError);
        }
      } else {
        notifyRoomConnectFail(NERoomConnectFailReason.rtcInitError);
      }
    }
  }

  void handleRaiseHandAction(HandsUpAction action) {
    roomData.userId2User[action.operateUser]?.raiseHandDetail =
        _InRoomRaiseHandDetail(
      action.subType == HandsUpSubType.hostRejectAudioHandsUp
          ? NEHandsUpStatus.downByHost
          : action.handsUp.status,
      action.handsUp.handsUpTime,
    );
  }

  Future<NEResult<MeetingInfo>> fetchRoomData([int? rtcId]) async {
    commonLogger.i('fetchRoomData: $rtcId');
    final result = await InRoomRepository.memberInfo(getCurrentRoomId(), avRoomUid: rtcId);
    if (result.isSuccess()) {
      final data = result.nonNullData;

      final roomInfo = roomData.roomInfo ?? createRoomInfo(data);
      roomData.roomInfo = roomInfo;
      roomInfo
        ..audioAllMute = data.audioAllMute == 1
        ..locked = JoinControlType.isLock(data.joinControlType)
        ..hostUserId = data.hostAccountId
        ..pinnedUserId = data.focusAccountId
        ..screenSharingUserId = data.screenSharersAccountId.isNotEmpty
            ? data.screenSharersAccountId.first
            : null
        ..settings = data.settings
        ..liveStreamInfo =
            data.live != null ? _InRoomLiveInfo(roomData, data.live!) : null;

      String? whiteboardShareUserId;
      if (data.whiteboardAvRoomUid.isNotEmpty) {
        //先从当前用户列表中查找
        //如果是首次同步数据，用户列表可能为空，需再次从返回数据的 member 列表中查询
        whiteboardShareUserId = roomData.rtcId2User[int.tryParse(data.whiteboardAvRoomUid.first)]?.userId
            ?? data.members.firstWhere((member) => member.avRoomUid.toString() == data.whiteboardAvRoomUid.first).accountId;
      }
      roomInfo.whiteboardSharingUserInfo = _WhiteboardSharingUserInfo(
        sharerUserId: whiteboardShareUserId,
        sharerImAccId: whiteboardShareUserId != null
            ? data.whiteboardOwnerImAccid.first
            : null,
        sharerRtcId: whiteboardShareUserId != null
            ? int.tryParse(data.whiteboardAvRoomUid.first)
            : null,
        whiteboardShareStatus: whiteboardShareUserId != null
            ? NERoomWhiteBoardShareStatus.start
            : NERoomWhiteBoardShareStatus.end,
      );

      roomData.userId2User.forEach((_, user) => updateUserWithRoomInfo(user));
      data.members.forEach(createOrUpdateInRoomUser);
    }
    return result;
  }

  _RoomInfo createRoomInfo(MeetingInfo data) {
    final info = _RoomInfo(
      uniqueId: data.meetingUniqueId,
      roomId: data.meetingId,
      shortRoomId: data.shortMeetingId,
      sipCid: data.sipCid,
      type: data.type,
      subject: data.subject,
      password: data.password,
      scheduleStartTime: data.scheduleStartTime,
      scheduleEndTime: data.scheduleEndTime,
      liveStreamInfo: data.live != null ? _InRoomLiveInfo(roomData, data.live!) : null,
      userId2User: roomData.userId2User,
      rtcId2User: roomData.rtcId2User,
    );

    info.audioAllMuteNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        commonLogger.i('onRoomSilentModeChanged: ${info.audioAllMute}');
        listeners.forEach((element) {
          element.onRoomSilentModeChanged(info.audioAllMute);
        });
      });
    info.hostUserIdNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        commonLogger.i('onRoomHostChanged: ${info.hostUserId}');
        /// 主持人发生变更，同步一次RoomInfo，因为主持人的音视频状态可能发生变更
        if (info.hostUserIdNotifier.before != info.hostUserIdNotifier.current) {
          fetchRoomData(getRtcId(info.hostUserIdNotifier.before));
          fetchRoomData(getRtcId(info.hostUserIdNotifier.current));
        }
        listeners.forEach((element) {
          element.onRoomHostChanged(info.hostUserId);
        });
      });
    info.lockedNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        final locked = roomData.roomInfo?.locked == true;
        commonLogger.i('onRoomLockStatusChanged: $locked');
        listeners.forEach((element) {
          element.onRoomLockStatusChanged(locked);
        });
      });
    info.liveStreamInfoNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        if (info.liveStreamInfo != null) {
          listeners.forEach((element) {
            commonLogger.i('onRoomLiveStreamInfoChanged');
            element.onRoomLiveStreamInfoChanged(info.liveStreamInfo!.copy());
          });
        }
      });
    Listenable.merge([
      info.pinnedUserIdNotifier,
      info.screenSharingUserIdNotifier,
      info.whiteboardSharingUserInfoNotifier,
    ]).addListener(() {
      roomData.userId2User.forEach((_, user) {
        updateUserWithRoomInfo(user);
      });
    });

    return info;
  }

  void updateRoomInfoOnUserLeave(String userId) {
    final roomInfo = roomData.roomInfo;
    if (roomInfo != null) {
      if (roomInfo.pinnedUserId == userId) {
        roomInfo.pinnedUserId = null;
      }
      if (roomInfo.screenSharingUserId == userId) {
        roomInfo.screenSharingUserId = null;
      }
      if (roomInfo.whiteboardSharingUserId == userId) {
        roomInfo.whiteboardSharingUserInfo = _WhiteboardSharingUserInfo(
            whiteboardShareStatus: NERoomWhiteBoardShareStatus.end);
      }
    }
  }

  _InRoomUser createOrUpdateInRoomUser(InMeetingMemberInfo memberInfo) {
    final videoStatus = indexToVideoStatus(memberInfo.video);
    final audioStatus = indexToAudioStatus(memberInfo.audio);
    final raiseHandDetail = memberInfo.handsUp
        .where((e) => e.handsUpType == HandsUpType.muteAll)
        .map((e) => _InRoomRaiseHandDetail(e.status, e.handsUpTime, ))
        .firstWhere((element) => true, orElse: () => _InRoomRaiseHandDetail.down());
    assert(roomData.roomInfo != null);
    final roomInfo = roomData.roomInfo!;
    final isPinned = memberInfo.accountId == roomInfo.pinnedUserId;
    final isScreenSharing = memberInfo.accountId == roomInfo.screenSharingUserId;
    final whiteBoardShareStatus =
        memberInfo.accountId == roomInfo.whiteboardSharingUserId
            ? NERoomWhiteBoardShareStatus.start
            : NERoomWhiteBoardShareStatus.end;
    final isOnline;
    if (isMyself(memberInfo.accountId)) {
      isOnline = isStateAtLeast(connected);
    } else {
      isOnline = memberInfo.isOnline;
    }

    final _InRoomUser user;
    if (roomData.userId2User.containsKey(memberInfo.accountId)) {
      user = roomData.userId2User[memberInfo.accountId]!;
      user.displayName = memberInfo.nickName;
      user.videoStatus = videoStatus;
      user.audioStatus = audioStatus;
      user.whiteBoardInteractStatus = _WhiteBoardInteractionStatusExtension.get(
          memberInfo.whiteBoardInteract);
      user.isOnline = isOnline;
      user.raiseHandDetail = raiseHandDetail;
      user.isPinned = user.userId == roomInfo.pinnedUserId;
      user.isScreenSharing = user.userId == roomInfo.screenSharingUserId;
      user.whiteBoardShareStatus = whiteBoardShareStatus;
    } else {
      user = createUser(
        memberInfo,
        videoStatus,
        audioStatus,
        isPinned,
        isScreenSharing,
        whiteBoardShareStatus,
        raiseHandDetail,
      );
    }
    return user;
  }

  _InRoomUser createUser(
    InMeetingMemberInfo memberInfo,
    NERoomVideoStatus videoStatus,
    NERoomAudioStatus audioStatus,
    bool isPinned,
    bool isScreenSharing,
    NERoomWhiteBoardShareStatus whiteBoardShareStatus,
    _InRoomRaiseHandDetail raiseHandDetail,
  ) {
    var initialAudioStatus = audioStatus;
    var initialVideoStatus = videoStatus;
    if (isMyself(memberInfo.accountId)) {
      initialAudioStatus = options.noAudio ? NERoomAudioStatus.off : NERoomAudioStatus.on;
      initialVideoStatus = options.noVideo ? NERoomVideoStatus.off : NERoomVideoStatus.on;
    }

    final user = _InRoomUser(
      userId: memberInfo.accountId,
      rtcId: memberInfo.avRoomUid,
      clientType: memberInfo.clientType,
      roleType: memberInfo.roleType,
      displayName: memberInfo.nickName,
      tag: memberInfo.memberTag,
      videoStatus: initialVideoStatus,
      audioStatus: initialAudioStatus,
      isPinned: isPinned,
      isScreenSharing: isScreenSharing,
      whiteBoardShareStatus: whiteBoardShareStatus,
      whiteBoardInteractStatus: _WhiteBoardInteractionStatusExtension.get(
          memberInfo.whiteBoardInteract),
      raiseHandDetail: raiseHandDetail,
      isOnline: memberInfo.isOnline,
    );

    user.displayNameNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        commonLogger.i('onRoomUserNameChanged: ${user.userId} ${user.displayName} ${user.displayName}');
        if (isMyself(user.userId)) {
          UserProfile.displayName = user.displayName;
          chatController.updateChatRoomUserRole(NEChatRoomUserRole(nick: user.displayName));
        }
        listeners.forEach((element) {
          element.onRoomUserNameChanged(user.userId, user.displayName);
        });
      });

    user.audioStatusNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        commonLogger.i('onRoomUserAudioStatusChanged: ${user.userId} ${user.displayName} ${user.audioStatus}');
        if (isMyself(user.userId)) {
          audioController.syncLocalAudioStatus();
        }
        if (!suppressAudioStatusChangeNotify) {
          listeners.forEach((element) {
            element.onRoomUserAudioStatusChanged(user.userId, user.audioStatus);
          });
        }
    });

    user.videoStatusNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        commonLogger.i('onRoomUserVideoStatusChanged: ${user.userId} ${user.displayName} ${user.videoStatus}');
        if (isMyself(user.userId)) {
          videoController.syncLocalVideoStatus();
        }
        listeners.forEach((element) {
          element.onRoomUserVideoStatusChanged(user.userId, user.videoStatus);
        });
    });

    user.raiseHandDetailNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        commonLogger.i('onRoomRaiseHandStatusChanged: ${user.userId} ${user.displayName} '
            'status=${user.raiseHandDetail}');
        listeners.forEach((element) {
          element.onRoomRaiseHandStatusChanged(user.userId, user.raiseHandDetail);
        });
    });

    user.pinnedNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        commonLogger.i('onRoomUserVideoPinStatusChanged: ${user.userId} ${user.displayName} ${user.isPinned}');
        listeners.forEach((element) {
          element.onRoomUserVideoPinStatusChanged(user.userId, user.isPinned);
        });
    });

    user.screenSharingNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        final status = user.isScreenSharing ? NERoomScreenShareStatus.start : NERoomScreenShareStatus.end;
        commonLogger.i('onRoomUserScreenShareStatusChanged: ${user.userId} ${user.displayName} $status');
        if(user.isScreenSharing) fetchRoomData(getRtcId(user.userId));
        listeners.forEach((element) {
          element.onRoomUserScreenShareStatusChanged(user.userId, status);
        });
    });

    user.whiteBoardSharingNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        commonLogger.i('onRoomUserWhiteBoardShareStatusChanged: ${user.userId} ${user.displayName} ${user.whiteBoardShareStatus}');
        listeners.forEach((element) {
          element.onRoomUserWhiteBoardShareStatusChanged(
              user.userId, user.whiteBoardShareStatus);
        });
    });

    user.whiteBoardInteractionStatusNotifier
      ..setNotifyDecider(changeNotifyDecider)
      ..addListener(() {
        commonLogger.i(
            'onWhiteBoardInteractionStatusChange: ${user.userId} ${user.displayName} ${user.whiteBoardInteractStatus}');
        if (isMyself(user.userId)) {
          whiteboardController.enableDraw(user.whiteBoardInteractStatus == NERoomWhiteBoardInteractionStatus.open);
        }
        listeners.forEach((element) {
          element.onRoomUserWhiteBoardInteractionStatusChanged(
              user.userId,
              user.whiteBoardInteractStatus ==
                  NERoomWhiteBoardInteractionStatus.open);
        });
    });

    roomData.userId2User[user.userId] = user;
    roomData.rtcId2User[user.rtcId] = user;
    if (isMyself(user.userId)) {
      roomData.self = user;
      user.videoStatus = videoStatus;
      user.audioStatus = audioStatus;
    }
    return user;
  }

  void updateUserWithRoomInfo(_InRoomUser user) {
    assert(roomData.roomInfo != null);
    final roomInfo = roomData.roomInfo!;
    user
      ..isPinned = user.userId == roomInfo.pinnedUserId
      ..isScreenSharing = user.userId == roomInfo.screenSharingUserId;

    final sharingUserInfo = roomInfo.whiteboardSharingUserInfoNotifier.value;
    if (user.whiteBoardInteractStatus == NERoomWhiteBoardInteractionStatus.open
        && sharingUserInfo.whiteboardShareStatus != NERoomWhiteBoardShareStatus.start) {
      user.whiteBoardInteractStatus = NERoomWhiteBoardInteractionStatus.close;
    }
    if (user.userId == sharingUserInfo.sharerUserId || user.isWhiteBoardSharing) {
      user.whiteBoardShareStatus = sharingUserInfo.whiteboardShareStatus;
    }
  }

  void scheduleEnsureUserOnlineTask() async {
    if (roomData.roomInfo == null) return;
    // wait more pending user
    await Future.delayed(Duration(milliseconds: 200));
    while (isActive && roomData.pendingUser.isNotEmpty) {
      // check and notify ready users
      final readyList = roomData.rtcId2User.values
          .where((user) => roomData.pendingUser.contains(user.rtcId) && user.isOnline)
          .toList();
      if (readyList.isNotEmpty) {
        roomData.pendingUser.removeAll(readyList.map((e) => e.rtcId));
        listeners.forEach((element) {
          commonLogger.i('onRoomUserJoin: count=${readyList.length}');
          element.onRoomUserJoin(readyList);
        });
      }

      if (roomData.pendingUser.isEmpty) return;
      // wait more pending user
      await Future.delayed(Duration(milliseconds: 200));

      await fetchRoomData(roomData.pendingUser.length >= 50 ? null : roomData.pendingUser.first);
    }
  }

  void userLeaveChannel(int rtcId) async {
    roomData.pendingUser.remove(rtcId);
    final user = roomData.rtcId2User[rtcId];
    if (user == null) return;
    updateRoomInfoOnUserLeave(user.userId);
    roomData.rtcId2User.remove(rtcId);
    roomData.userId2User.remove(user.userId);
    listeners.forEach((element) {
      commonLogger.i('onRoomUserLeave: ${user.userId} ${user.displayName}');
      element.onRoomUserLeave([user]);
    });
  }

  @override
  void onAudioHasHowling() {}

  @override
  void onClientRoleChange(int oldRole, int newRole) {}

  @override
  void onConnectionStateChanged(int state, int reason) {
    commonLogger.i('onConnectionStateChanged state=$state reason=$reason');
  }

  @override
  void onConnectionTypeChanged(int newConnectionType) {}

  @override
  void onDisconnect(int rtcReason) {
    commonLogger.i('onDisconnectFromRtc: reason=$rtcReason');
    var reason = NERoomDisconnectReason.rtcSDKError;
    switch (rtcReason) {
      ///房间被关闭
      case NERtcErrorCode.roomClosed:
        reason = isMySelfHost()
            ? NERoomDisconnectReason.closeBySelf
            : NERoomDisconnectReason.closeByHost;
        break;
      ///被管理员踢出房间
      case NERtcErrorCode.serverKicked:
        reason = NERoomDisconnectReason.removedByHost;
        break;
      case NERtcErrorCode.ok:
        return;
    }
    notifyRoomDisconnected(reason);
  }

  @override
  void onError(int code) {}

  @override
  void onFirstAudioDataReceived(int uid) {}

  @override
  void onFirstAudioFrameDecoded(int uid) {}

  @override
  void onFirstVideoDataReceived(int uid) {
    commonLogger.i('firstVideoFrameTime rtcId=$uid');
    trackPeriodicEvent(TrackEventName.firstVideoDataReceived, extra: {
      'member_uid': uid,
      'meeting_id': roomData.joinRoomInfo.avRoomCid
    });
  }

  @override
  void onFirstVideoFrameDecoded(int uid, int width, int height) {}

  @override
  void onJoinChannel(int result, int channelId, int elapsed) {
    if (!isActive) {
      return;
    }
    final isChannelIdMatch = channelId.toString() == joinRoomInfo.avRoomCid;
    commonLogger.i('onJoinChannel: result=$result channelId=$channelId'
        ' elapsed=$elapsed channelIdMatch=$isChannelIdMatch');
    if (result == NERtcErrorCode.ok && isChannelIdMatch) {
      trackPeriodicEvent(TrackEventName.joinChannelSuccess);
      state = connected;
      lifecycleOwner.notifyRoomOnConnected();
      roomData.self.isOnline = true;
      changeNotifyDecider.shouldNotify = true;
      audioController.setUp();
      videoController.setUp();
    } else {
      trackPeriodicEvent(TrackEventName.joinChannelFailed,
          extra: {'value': result});
      notifyRoomConnectFail(NERoomConnectFailReason.joinChannelError);
    }
  }

  void onRtcReportUserAudioStatus(int rtcId, bool started) {
    if (!isActive) return;
    final user = roomData.rtcId2User[rtcId];
    if (user != null) {
      user.audioStartedOnRtcSide = started;
      if (user.audioStatusInconsistentWithRtc) {
        scheduleCheckAVStatusConsistencyTask();
      }
    }
  }

  void onRtcReportUserVideoStatus(int rtcId, bool started) {
    if (!isActive) return;
    final user = roomData.rtcId2User[rtcId];
    if (user != null) {
      user.videoStartedOnRtcSide = started;
      if (user.videoStatusInconsistentWithRtc) {
        scheduleCheckAVStatusConsistencyTask();
      }
    }
  }

  void scheduleCheckAVStatusConsistencyTask() {
    checkAVStatusConsistencyTask ??= Timer(Duration(seconds: 3), () {
      if (!isActive) return;
      commonLogger.i('check av status consistency task started');
      roomData.userId2User.values
          .toList(growable: false)
          .where((user) =>
              user.audioStatusInconsistentWithRtc ||
              user.videoStatusInconsistentWithRtc)
          .forEach((user) {
            fetchRoomData(user.rtcId);
          });
      checkAVStatusConsistencyTask = null;
    });
  }

  ///********************nrtc callback******************************///

  @override
  void onLeaveChannel(int result) {
    commonLogger.i('onLeaveChannel result=$result');
  }

  @override
  void onLiveStreamState(String taskId, String pushUrl, int liveState) {
    commonLogger.i('onLiveStreamState taskId=$taskId pushUrl=$pushUrl liveState=$liveState');
  }

  @override
  void onLocalAudioVolumeIndication(int volume) {}

  @override
  void onReJoinChannel(int result) {
    if (!isActive) {
      return;
    }
    commonLogger.i('onReJoinChannel result=$result');
    if (result == NERtcErrorCode.ok) {
      fetchRoomData();
    } else {
      notifyRoomDisconnected(NERoomDisconnectReason.rejoinChannelError);
    }
  }

  @override
  void onReceiveSEIMsg(int userID, String seiMsg) {}

  @override
  void onReconnectingStart() {}

  @override
  void onRemoteAudioVolumeIndication(
      List<NERtcAudioVolumeInfo> volumeList, int totalVolume) {
    if (!isActive) {
      return;
    }
    final volumeInfo = volumeList
        .where((info) =>
            info.uid != null &&
            info.volume != null &&
            roomData.rtcId2User[info.uid] != null)
        .map((info) => NERoomUserAudioVolumeInfo(
            roomData.rtcId2User[info.uid]!.userId, info.volume!))
        .toList();
    listeners.forEach((element) {
      element.onRoomRemoteAudioVolumeIndication(
        volumeInfo,
        totalVolume,
      );
    });
  }

  @override
  void onUserAudioMute(int uid, bool muted) {}

  @override
  void onUserAudioStart(int rtcId) {
    commonLogger.i('onUserAudioStart: $rtcId');
    onRtcReportUserAudioStatus(rtcId, true);
  }

  @override
  void onUserAudioStop(int rtcId) {
    commonLogger.i('onUserAudioStop: $rtcId');
    onRtcReportUserAudioStatus(rtcId, false);
  }

  @override
  void onUserJoined(int rtcId) {
    if (!isActive) {
      return;
    }
    commonLogger.i('onUserJoined: $rtcId');
    roomData.pendingUser.add(rtcId);
    if (roomData.pendingUser.length == 1) {
      scheduleEnsureUserOnlineTask();
    }
  }

  @override
  void onUserLeave(int rtcId, int reason) {
    if (!isActive) {
      return;
    }
    commonLogger.i('onUserLeave: $rtcId $reason');
    userLeaveChannel(rtcId);
  }

  @override
  void onUserSubStreamVideoStart(int rtcId, int maxProfile) {
    commonLogger.i('onUserSubStreamVideoStart: $rtcId $maxProfile');
  }

  @override
  void onUserSubStreamVideoStop(int rtcId) {
    commonLogger.i('onUserSubStreamVideoStop: $rtcId');
  }

  @override
  void onUserVideoMute(int uid, bool muted) {}

  @override
  void onUserVideoProfileUpdate(int uid, int maxProfile) {
    trackPeriodicEvent(TrackEventName.memberChangeProfile, extra: {
      'member_profile': maxProfile,
      'member_uid': uid,
      'meeting_id': roomData.joinRoomInfo.meetingId,
    });
  }

  @override
  void onUserVideoStart(int rtcId, int maxProfile) {
    commonLogger.i('onUserVideoStart: $rtcId $maxProfile');
    onRtcReportUserVideoStatus(rtcId, true);
  }

  @override
  void onUserVideoStop(int rtcId) {
    commonLogger.i('onUserVideoStop: $rtcId');
    onRtcReportUserVideoStatus(rtcId, false);
  }

  @override
  void onWarning(int code) {
    commonLogger.i('onWarning: $code');
  }

  @override
  void onAudioRecording(int code, String filePath) {

  }

  @override
  void onMediaRelayReceiveEvent(int event, int code, String channelName) {

  }

  @override
  void onMediaRelayStatesChange(int state, String channelName) {

  }

  @override
  void onLocalPublishFallbackToAudioOnly(bool isFallback, int streamType) {

  }

  @override
  void onRemoteSubscribeFallbackToAudioOnly(int uid, bool isFallback, int streamType) {

  }

  void release() {
    didFinish();
    screenShareController.release();
    beautyController.release();
    chatController.release();
    whiteboardController.release();

    ///释放G2
    NERoomMediaService.instance.release(MediaType.inRoom);
  }


  @override
  void onRemoteVideoStats(List<NERtcVideoRecvStats> statsList) {
    [...rtcStatsEventListeners].forEach((element) {
      var newStatsList = <NERoomVideoRecvStats>[];
      statsList.forEach((stats) {
        var userId = roomData.rtcId2User[stats.uid ?? 0]?.userId;
        if(userId !=null) {
          newStatsList.add(NERoomVideoRecvStats(
            userId: userId,
            layers: stats.layers,
          ));
        }
      });
      element.onRemoteVideoStats(newStatsList);
    });
  }

  @override
  void onLocalAudioStats(NERtcAudioSendStats stats) {
    [...rtcStatsEventListeners].forEach((element) {
      element.onLocalAudioStats(stats);
    });
  }

  @override
  void onRemoteAudioStats(List<NERtcAudioRecvStats> statsList) {
    [...rtcStatsEventListeners].forEach((element) {
      var newStatsList = <NERoomAudioRecvStats>[];
      statsList.forEach((stats) {
        var userId = roomData.rtcId2User[stats.uid ?? 0]?.userId;
        if(userId !=null) {
          newStatsList.add(NERoomAudioRecvStats(
            userId: userId,
            kbps: stats.kbps,
            lossRate: stats.lossRate,
            volume: stats.volume,
            totalFrozenTime: stats.totalFrozenTime,
            frozenRate: stats.frozenRate,
          ));
        }
      });
      element.onRemoteAudioStats(newStatsList);
    });
  }

  @override
  void onRtcStats(NERtcStats stats) {
    [...rtcStatsEventListeners].forEach((element) {
      element.onRtcStats(stats);
    });
  }

  @override
  void onLocalVideoStats(NERtcVideoSendStats stats) {
    [...rtcStatsEventListeners].forEach((element) {
      element.onLocalVideoStats(stats);
    });
  }

  @override
  void onNetworkQuality(List<NERtcNetworkQualityInfo> statsList) {
    [...rtcStatsEventListeners].forEach((element) {
      var newNetworkQualityInfo = <NERoomNetworkQualityInfo>[];
      statsList.forEach((stats) {
        var userId = roomData.rtcId2User[stats.uid ?? 0]?.userId;
        if(userId !=null){
          newNetworkQualityInfo.add(NERoomNetworkQualityInfo(
            userId: userId,
            txQuality: stats.txQuality,
            rxQuality: stats.rxQuality,
          ));
        }
      });
      element.onNetworkQuality(newNetworkQualityInfo);
    });
  }

  /// [NERtcAudioDevice]
  @override
  void onAudioDeviceChanged(int selected) {
    commonLogger.i('onAudioDeviceChanged $selected');
    [...rtcDeviceEventListeners].forEach((element) {
      element.onAudioDeviceChanged(selected);
    });
  }

  @override
  void onAudioDeviceStateChange(int deviceType, int deviceState) {
    commonLogger.i('onAudioDeviceStateChange: $deviceType $deviceState');
    [...rtcDeviceEventListeners].forEach((element) {
      element.onAudioDeviceStateChange(deviceType, deviceState);
    });
  }

  @override
  void onVideoDeviceStageChange(int deviceState) {
    commonLogger.i('onVideoDeviceStageChange: $deviceState');
    [...rtcDeviceEventListeners].forEach((element) {
      element.onVideoDeviceStageChange(deviceState);
    });
  }
}

class _RoomInfoChangeNotifyDecider extends NotifyDecider {

  bool _value = false;

  bool get shouldNotify => _value;

  set shouldNotify(bool value) {
    if (_value != value) {
      _value = value;
      notifyListeners();
    }
  }

  @override
  bool updateShouldNotify() => shouldNotify;
}
