// Copyright (c) 2014-2021 NetEase, Inc.
// All right reserved.

part of room_kit;

/// 媒体服务
/// TODO接口需要拆分，设置进来
class NERoomMediaService
    with
        _RoomAloggerMixin,
        NERtcVideoRendererEventListener {

  factory NERoomMediaService() => _getInstance();

  static NERoomMediaService get instance => _getInstance();

  static NERoomMediaService? _instance;

  static NERoomMediaService _getInstance() {
    _instance ??= NERoomMediaService._internal();
    return _instance!;
  }

  late  NERtcEngine _engine;

  late var _isInit = false;

  MediaType _currentMediaType = MediaType.none;
  ///engine
  NERtcEngine get engine => _engine;

  NERoomMediaService._internal() {
    _engine = NERtcEngine();
  }

  Future<int> initRtcEngine({
    required NERtcChannelEventCallback rtcChannelEventCallback,
    bool attendeeRecordOn = false,
    required MediaType mediaType,
    required String nrtcAppKey,
  }) async {
    if (_isInit) {
      if (_currentMediaType == mediaType) {
        return NERtcErrorCode.ok;
      } else {
        if (_currentMediaType == MediaType.inRoom) return NERtcErrorCode.ok;
        await engine.release();
        _isInit = !_isInit;
      }
    }

    var options = getOption(attendeeRecordOn);
    var result = await engine.create(
      appKey:nrtcAppKey,
      channelEventCallback: rtcChannelEventCallback,
      options: options,
    );
    _isInit = result == NERtcErrorCode.ok;
    _currentMediaType = mediaType;
    return result;
  }

  /// join channel
  Future<int> joinChannel(JoinRoomInfo joinRoomInfo) async {
    var result = await engine.joinChannel(joinRoomInfo.avRoomCheckSum,
        joinRoomInfo.avRoomCName, joinRoomInfo.avRoomUid);
    commonLogger.i('joinChannel result=$result');
    return result;
  }

  Future leaveChannel() async {
    final result = await engine.leaveChannel();
    commonLogger.i('leaveChannel result=$result');
  }

  @override
  void onFrameResolutionChanged(int uid, int width, int height, int rotation) {
    // TODO: implement onFrameResolutionChanged
  }

  /// 会议录制模式，0：混合与单人，1：混合，2：单人，但只在有值的时候，才会由服务器抄送给应用的回调接口
  NERtcServerRecordMode get meetingRecordMode {
    var mode = SettingsRepository().meetingRecordMode;
    if (mode >= NERtcServerRecordMode.values.length || mode < 0) {
      mode = 0;
    }
    return NERtcServerRecordMode.values[mode];
  }

  NERtcOptions getOption(bool attendeeRecordOn) {
    return NERtcOptions(
        logLevel: NERtcLogLevel.info,
        logDir: NERoomLogService()._rtcSDKPath,
        audioAutoSubscribe: false,
        videoEncodeMode: useHardwareCodec(),
        videoDecodeMode: useHardwareCodec(),
        serverRecordAudio: SettingsRepository().meetingRecordAudioEnable &&
            attendeeRecordOn &&
            SettingsRepository.isMeetingCloudRecordSupported(),
        serverRecordVideo: SettingsRepository().meetingRecordVideoEnable &&
            attendeeRecordOn &&
            SettingsRepository.isMeetingCloudRecordSupported(),
        serverRecordMode: meetingRecordMode,
        publishSelfStream: true,
        videoCaptureObserverEnabled: true,
        /// IOS 平台美颜回调
        videoSendMode: NERtcVideoSendMode.none,
        serverRecordSpeaker: true);
  }

  @override
  void onFirstFrameRendered(int uid) {
    // TODO: implement onFirstFrameRendered
  }

  Future<void> setConfig() async {
    await engine.setAudioProfile(
        NERtcAudioProfile.profileDefault, NERtcAudioScenario.scenarioSpeech);
    var config = NERtcVideoConfig()..videoProfile = NERtcVideoProfile.hd720p;
    await engine.setLocalVideoConfig(config);
  }

  NERtcMediaCodecMode useHardwareCodec() {
    return defaultTargetPlatform == TargetPlatform.iOS
        ? NERtcMediaCodecMode.hardware
        : NERtcMediaCodecMode.software;
  }

  ///美颜预览配置
  Future<NERtcVideoRenderer> createVideoPreview() async {
    var config = NERtcVideoConfig();
    config.videoProfile = NERtcVideoProfile.hd720p;
    config.frameRate = NERtcVideoFrameRate.fps_30;
    config.degradationPrefer = NERtcDegradationPreference.degradationDefault;
    config.frontCamera = true;
    config.videoCropMode = NERtcVideoCropMode.cropDefault;
    await _engine.setLocalVideoConfig(config);
    var renderer = await VideoRendererFactory.createVideoRenderer();
    await renderer.setMirror(true);
    await renderer.attachToLocalVideo();
    await _engine.enableLocalVideo(true);
    await _engine.startVideoPreview();
    return renderer;
  }

  Future<void> release(MediaType mediaType) async {
    if (!_isInit || mediaType != _currentMediaType) return;
    _isInit = false;
    _currentMediaType = MediaType.none;
    await _engine.leaveChannel();
    await _engine.enableLocalVideo(false);
    await _engine.enableLocalAudio(false);
    await _engine.stopVideoPreview();
    await _engine.release();
  }
}
enum MediaType { none,preRoom, inRoom }

class NERoomAudioRecvStats {
  /// 用户 ID，指定是哪个用户的音频流
   String  userId ;
   /// 接收到的码率 (kbps)
   int? kbps;

   /// 丢包率
   int? lossRate;

   /// 音量[0-100]
   int? volume;

   /// 音频卡顿累计时长，从收到远端用户音频算起
   int? totalFrozenTime;

   /// 平均音频卡顿率
   int? frozenRate;

   NERoomAudioRecvStats({required this.userId, this.kbps, this.lossRate, this.volume,
     this.totalFrozenTime, this.frozenRate});

   @override
   String toString() {
     return 'NERoomAudioRecvStats{userId: $userId, kbps: $kbps, lossRate: $lossRate, volume: $volume, totalFrozenTime: $totalFrozenTime, frozenRate: $frozenRate}';
   }

}

/// 用户的网络质量
class NERoomNetworkQualityInfo {
  /// 用户 ID
  String userId;

  /// 上行网络质量 [NERTcNetworkStatus]
  int? txQuality;

  /// 下行网络质量 [NERTcNetworkStatus]
  int? rxQuality;

  NERoomNetworkQualityInfo({required this.userId, this.txQuality, this.rxQuality});

  @override
  String toString() {
    return 'NERoomNetworkQualityInfo{userId: $userId, txQuality: $txQuality, rxQuality: $rxQuality}';
  }
}

/// 远端视频流的统计信息
class NERoomVideoRecvStats {
  /// 用户 ID，指定是哪个用户的视频流
  String userId;

  /// 当前uid 每条流的接收下行统计信息
  List<NERtcVideoLayerRecvStats> layers = <NERtcVideoLayerRecvStats>[];


  NERoomVideoRecvStats({required this.userId, required this.layers});

  @override
  String toString() {
    return 'NERoomVideoRecvStats{userId: $userId, layers: $layers}';
  }
}



