// Copyright (c) 2014-2020 NetEase, Inc.
// All right reserved.
part of room_kit;

class _NERoomServiceImpl extends NERoomService with EventTrackMixin {
  static const _tag = 'NERoomServiceImpl';

  static final _NERoomServiceImpl _instance = _NERoomServiceImpl._();

  factory _NERoomServiceImpl() => _instance;

  _NEInRoomServiceImpl? currentInRoomService;

  _NERoomServiceImpl._();
  @override
  Future<NEResult<JoinRoomInfo>> joinRoom(
      NEJoinRoomParams param, NEJoinRoomOptions? opts) async {
    Alog.i(
        tag: _tag,
        moduleName: _moduleName,
        type: AlogType.api,
        content: 'joinRoom, param$param, opts$opts');
    var checkConnectivity = await Connectivity().checkConnectivity();
    if (checkConnectivity == ConnectivityResult.none) {
      return NEResult(code: RoomErrorCode.networkError);
    }
    if (param.roomId?.isEmpty ?? true) {
      return NEResult(code: RoomErrorCode.paramError, msg: 'Empty roomId');
    }
    return await _joinRoom(param, opts);
  }

  Future<NEResult<JoinRoomInfo>> _joinRoom(
      NEJoinRoomParams param, NEJoinRoomOptions? opts) async {
    ///创建默认opts
     opts ??= NEJoinRoomOptions(noAudio: true,noVideo: true);
    var check = await _stateError();
    if (check.code != RoomErrorCode.success) return Future.value(check);

    Future<NEResult<JoinRoomInfo>> onSuccess() async {
      NEResult<JoinRoomInfo> _joinRoomInfo;
      bool anonymous;
      if (AuthRepository.isAuthorized()) {
        UserProfile.isAnonymous = false;
        anonymous = false;
        _joinRoomInfo = await _normalJoinRoom(param, opts);
      } else {
        UserProfile.isAnonymous = true;
        anonymous = true;
        _joinRoomInfo = await _anonymousJoinRoom(param, opts);
      }
      if (_joinRoomInfo.isSuccess()) {
        ensureInRoomService(anonymous, _joinRoomInfo.nonNullData, opts!);
      }
      return _joinRoomInfo;
    }

    return InvocationManager().asyncWithResource(
      [Resource.account, Resource.room],
      onSuccess,
      onJoinRoomFail,
    );
  }

  Future<NEResult<JoinRoomInfo>> _anonymousJoinRoom(
      NEJoinRoomParams param, NEJoinRoomOptions? opts) async {
    trackPeriodicEvent(TrackEventName.roomJoin,
        extra: {'value': _TrackExtra.joinRoomAnonymous});
    var anonymousJoinRoomInfo = await RoomRepository.anonymousJoinRoom(
      param.roomId!,
      param.displayName,
      param.tag,
      opts == null
          ? AVState.close
          : opts.noVideo
              ? AVState.close
              : AVState.open,
      opts == null
          ? AVState.close
          : opts.noAudio
              ? AVState.close
              : AVState.open,
      param.password,
    );
    if (anonymousJoinRoomInfo.isSuccess()) {
      return NEResult(
        code: RoomErrorCode.success,
        data: anonymousJoinRoomInfo.data!.joinRoomInfo,
      );
    }
    return NEResult(
      code: anonymousJoinRoomInfo.code,
      msg: anonymousJoinRoomInfo.msg,
    );
  }

  Future<NEResult<JoinRoomInfo>> _normalJoinRoom(
      NEJoinRoomParams param, NEJoinRoomOptions? opts) {
    trackPeriodicEvent(TrackEventName.roomJoin,
        extra: {'value': _TrackExtra.joinRoomAuthorized});
    return RoomRepository.joinRoom(
        param.displayName,
        param.roomId!,
        param.tag,
        opts == null
            ? AVState.close
            : opts.noVideo
                ? AVState.close
                : AVState.open,
        opts == null
            ? AVState.close
            : opts.noAudio
                ? AVState.close
                : AVState.open,
        param.password);
  }

  @override
  Future<NEResult<JoinRoomInfo>> startRoom(
      NEStartRoomParams param, NEStartRoomOptions? opts) async {
    Alog.i(
        tag: _tag,
        moduleName: _moduleName,
        type: AlogType.api,
        content: 'startRoom, param$param, opts$opts');
    // if (!checkParameters(param, opts, listener)) return;
    ///创建默认opts
    opts ??= NEStartRoomOptions(noVideo: true, noAudio: true);
    var result = await Connectivity().checkConnectivity();
    if (result == ConnectivityResult.none) {
      return Future.value(
          NEResult(code: RoomErrorCode.networkError));
    }
    var startRoomResult = await _startRoom(param, opts);
    if (startRoomResult.isSuccess()) {
      ensureInRoomService(false, startRoomResult.nonNullData, opts);
    }
    return startRoomResult;
  }

  Future<NEResult<JoinRoomInfo>> _startRoom(
      NEStartRoomParams param, NEStartRoomOptions? opts) async {
    var check = await _stateError(checkAuth: true);
    if (check.code != RoomErrorCode.success) return Future.value(check);

    Future<NEResult<JoinRoomInfo>> onSuccess() async {
      trackPeriodicEvent(TrackEventName.roomCreate);
      var data = await RoomRepository.createRoom(
        param.displayName,
        param.roomId,
        param.password,
        param.tag,
        opts == null
            ? AVState.close
            : opts.noVideo
                ? AVState.close
                : AVState.open,
        opts == null
            ? AVState.close
            : opts.noAudio
                ? AVState.close
                : AVState.open,
        opts == null ? true : !opts.noChat,
        opts == null ? false : !opts.noCloudRecord,
        param.scene != null ? param.scene!.toJson() : null,
      );
      return trackResultEvent(
        data,
        TrackEventName.roomCreateSuccess,
        TrackEventName.roomCreateFailed,
      );
    }

    return InvocationManager().asyncWithResource(
      [Resource.account, Resource.room],
      onSuccess,
      onJoinRoomFail,
    );
  }

  void ensureInRoomService(bool anonymous, JoinRoomInfo joinInfo, NERoomOptions options) {
    final self = _NEInRoomServiceImpl(anonymous, joinInfo, options);
    currentInRoomService = self;
    self.finished.then((value) {
      Alog.i(
          tag: _tag,
          moduleName: _moduleName,
          content: 'room finished: roomId=${self.getCurrentRoomId()}');
      if (self == currentInRoomService) {
        currentInRoomService = null;
      }
    });
  }

  @override
  Future<NEResult<void>> leaveCurrentRoom(bool closeIfHost) async {
    Alog.i(
        tag: _tag,
        moduleName: _moduleName,
        content: 'leaveCurrentRoom: closeIfHost=$closeIfHost');
    return currentInRoomService?.leaveCurrentRoom(closeIfHost) ??
        Future.value(NEResult(code: RoomErrorCode.failed));
  }

  @override
  NERoomInfo? getCurrentRoomInfo() {
    return isCurrentRoomConnected()
        ? currentInRoomService?.roomData.roomInfo
        : null;
  }

  NEResult<JoinRoomInfo> trackResultEvent(
      NEResult<JoinRoomInfo> result, String success, String fail) {
    if (result.code == RoomErrorCode.success) {
      trackPeriodicEvent(success);
    } else {
      trackPeriodicEvent(fail, extra: {'value': result.code});
    }
    return result;
  }

  Future<NEResult<JoinRoomInfo>> _stateError({bool checkAuth = false}) {
    if (checkAuth && !AuthRepository.isAuthorized()) {
      return Future.value(NEResult(code: RoomErrorCode.notLogin));
    }
    // if (isCurrentRoomConnected()) {
    //   return Future.value(NEResult(code: RoomErrorCode.alreadyInRoom));
    // }
    return Future.value(NEResult(code: RoomErrorCode.success));
  }

  bool isCurrentRoomConnected() {
    return currentInRoomService?.isActive ?? false;
  }
}
