//
// Copyright (c) 2021 NetEase Inc
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal in
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
// the Software, and to permit persons to whom the Software is furnished to do so,
// subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

part of room_kit;

extension ALogLevelExtension on ALogLevel {
  static ALogLevel get(int? mode) {
    return ALogLevel.values.firstWhere(
      (element) => element.index == mode,
      orElse: () => ALogLevel.verbose,
    );
  }
}

class ALoggerConfig {
  final ALogLevel level;
  final String? path;
  final String? namePrefix;

  const ALoggerConfig(
      {this.level = ALogLevel.verbose, this.path, this.namePrefix});

  static ALoggerConfig? fromMap(Map? map) {
    if (map == null) {
      return null;
    }
    return ALoggerConfig(
      level: ALogLevelExtension.get(map['level'] as int?),
      path: map['path'] as String?,
      namePrefix: map['namePrefix'] as String?,
    );
  }

  @override
  String toString() {
    return 'ALoggerConfig{level: $level, path: $path, namePrefix: $namePrefix}';
  }
}

class NERoomLogService {
  static final NERoomLogService _instance = NERoomLogService._();

  NERoomLogService._();

  factory NERoomLogService() => _instance;

  String? _rootPath;

  ///_rootPath必须在init接口之后，才能获取日志path
  String get rootPath => _rootPath ?? '';

  String get _roomSDKPath => '${rootPath}RoomKit/';

  String get _rtcSDKPath => '${rootPath}NERtcSDK/';

  ///nim ios默认增加 [NIMSDK] ，此处增加平台处理
  String get _nimSDKPath => Platform.isIOS ? rootPath : '${rootPath}NIMSDK/';

  Future<bool> init({ALoggerConfig? loggerConfig}) async {
    if (_rootPath != null) {
      return true;
    }
    loggerConfig ??= ALoggerConfig();
    var logRootPath = loggerConfig.path;
    if (logRootPath?.isEmpty ?? true) {
      logRootPath = await _defaultLogRootPath;
    }
    //此处适配iOS，传入路径后缀无分隔符'/'；
    _rootPath = logRootPath!.endsWith('/') ? logRootPath : '$logRootPath/';
    if (!(await _createDirectory(rootPath))) return false;
    if (!(await _createDirectory(_nimSDKPath))) return false;
    if (!(await _createDirectory(_rtcSDKPath))) return false;

    final success = Alog.init(
        loggerConfig.level, _roomSDKPath, loggerConfig.namePrefix ?? 'roomkit');
    print('RoomLogService init with path: $rootPath, success: $success');
    if (!success) {
      _rootPath = null;
    }
    return success;
  }

  static Future<bool> _createDirectory(String path) async {
    var isCreate = false;
    var filePath = Directory(path);
    try {
      if (!await filePath.exists()) {
        await filePath.create();
        isCreate = true;
      } else {
        isCreate = true;
      }
    } catch (e) {
      isCreate = false;
      print('error $e');
    }
    return isCreate;
  }

  static Future<String> get _defaultLogRootPath async {
    var directory;
    if (Platform.isIOS) {
      directory = await getApplicationDocumentsDirectory();
    } else {
      directory = await getExternalStorageDirectory();
    }
    return '${directory.path}/log/';
  }
}
