// Copyright (c) 2014-2020 NetEase, Inc.
// All right reserved.
part of room_kit;

/// 提供创建和加入房间时必要的基本参数，如房间ID、房间昵称等
///
/// See also:
/// * [NEStartRoomParams]
/// * [NEJoinRoomParams]
class NERoomParams {
  /// 指定要创建或加入的目标房间ID
  /// * 加入房间时，该字段必须是一个当前正在进行中的房间ID，不能为空
  /// * 创建房间时，该字段可使用通过NEAccountService.getPersonalRoomId返回的个人房间号，或者不指定(置空)；
  /// 当不指定房间ID创建房间时，由服务端随机分配一个房间ID
  final String? roomId;

  /// 房间中的用户昵称，不能为空
  final String displayName;

  /// 房间密码; 
  /// 创建房间时，如果密码不为空，会创建带指定密码的房间
  /// 加入带密码的房间时，需要指定密码
  final String? password;

  /// 会议中的用户成员标签，自定义，最大长度50
  final String? tag;

  NERoomParams({
    this.roomId,
    required this.displayName,
    this.password,
    this.tag,
  });

  NERoomParams.fromMap(Map map)
      : this(
          roomId: map['roomId'] as String?,
          displayName: (map['displayName'] ?? '')as String,
          password: map['password'] as String?,
          tag: map['tag'] as String?,
        );

  @override
  String toString() {
    return 'NERoomParams{roomId: $roomId, displayName: $displayName, tag: $tag}';
  }
}

/// 提供创建房间时必要的额外参数，如房间ID、用户房间昵称等
class NEStartRoomParams extends NERoomParams {

  /// 房间场景字段
  final NERoomScene? scene;

  NEStartRoomParams({
    String? roomId,
    required String displayName,
    String? password,
    String? tag,
    this.scene,
  }) : super(
          roomId: roomId,
          displayName: displayName,
          password: password,
          tag: tag,
        );

  NEStartRoomParams.fromMap(Map map)
      : this(
          roomId: map['roomId'] as String?,
          displayName: (map['displayName'] ?? '') as String,
          password: map['password'] as String?,
          tag: map['tag'] as String?,
          scene: map['scene'] as NERoomScene?,
        );
}

/// 提供加入房间时必要的额外参数，如房间ID、用户房间昵称等
class NEJoinRoomParams extends NERoomParams {

  NEJoinRoomParams({
    required String roomId,
    required String displayName,
    String? password,
    String? tag,
  }) : super(
          roomId: roomId,
          displayName: displayName,
          password: password,
          tag: tag,
        );

  NEJoinRoomParams.fromMap(Map map)
      : super.fromMap(map);

  @override
  String toString() {
    return 'NEJoinRoomParams{password: $password}';
  }
}

/// 提供创建和加入房间时必要的基本配置信息和选项开关，通过这些配置和选项可控制入会时的行为，如音视频的开启状态等
///
/// See also:
/// * [NEStartRoomOptions]
/// * [NEJoinRoomOptions]
class NERoomOptions {
  /// 配置入会时是否关闭本端视频，默认为true，即关闭视频，但在房间中可重新打开
  final bool noVideo;

  /// 配置入会时是否关闭本端音频，默认为true，即关闭音频，但在房间中可重新打开
  final bool noAudio;

  /// 额外选项
  final Map<String, dynamic> extras;

  NERoomOptions({
    this.noVideo = true,
    this.noAudio = true,
    this.extras = const <String, dynamic>{},
  });

  @override
  String toString() {
    return 'NERoomOptions{noVideo: $noVideo, noAudio: $noAudio, extras: $extras}';
  }
}

/// 提供创建房间时必要的配置信息和选项开关，如音频、视频开关
class NEStartRoomOptions extends NERoomOptions {
  /// 配置房间中是否显示显示云端录制状态
  final bool noCloudRecord;

  /// 配置房间内是否开启聊天室功能
  final bool noChat;

  NEStartRoomOptions({
    bool? noVideo,
    bool? noAudio,
    this.noCloudRecord = true,
    this.noChat = false,
  }) : super(
          noVideo: noVideo ?? true,
          noAudio: noAudio ?? true,
        );

  @override
  String toString() {
    return 'NEStartRoomOptions{noCloudRecord: $noCloudRecord, noChat: $noChat}';
  }
}

/// 提供加入房间时必要的配置信息和选项开关，如音频、视频开关
class NEJoinRoomOptions extends NERoomOptions {
  NEJoinRoomOptions({
    bool? noVideo,
    bool? noAudio,
  }) : super(
          noVideo: noVideo ?? true,
          noAudio: noAudio ?? true,
        );
}

/// 生命周期回调接口，用于监听房间状态变更事件
abstract class NERoomLifeCycleObserver{
  ///Disconnected之后，状态默认重置回调
  void onIdle();
  ///startRoom && joinRoom 成功后进行回调
  void onConnecting();
  /// 加入房间失败，如：初始化 rtc 失败、加入频道失败
  /// [NERoomConnectFailReason]
  void onConnectFail(int reason);
  ///同步完房间内准备工作，在onJoinChannel回调之后则进行回调
  void onConnected();
  ///leaveRoom ||  quitRoom  || kickOut 触发时，则进行回调
  void onDisconnecting();
  ///leaveRoom ||  quitRoom || kickOut 完成后，则进行回调；[reason] 是回调Disconnected的原因
  /// [NERoomDisconnectReason]
  void onDisconnected(int reason);
}

/// 房间失败原因
class NERoomConnectFailReason {
  /// 房间不存在
  static const int roomNotExist = 1;
  /// 同步房间信息失败
  static const int syncDataError = 2;
  /// rtc 模块初始化失败
  static const int rtcInitError = 3;
  /// 加入频道失败
  static const int joinChannelError = 4;
}

/// 房间结束原因
class NERoomDisconnectReason {
  /// 用户主动离开房间
  static const int leaveBySelf = 1;

  /// 用户主动关闭房间
  static const int closeBySelf = 2;

  /// 被主持人移除房间
  static const int removedByHost = 3;

  /// 房间被主持人关闭
  static const int closeByHost = 4;

  /// 账号重复登录
  static const int loginOnOtherDevice = 5;

  /// 登录信息过期
  static const int authInfoExpired = 6;

  /// 重新加入频道失败
  static const int rejoinChannelError = 7;

  /// rtc sdk error
  static const int rtcSDKError = 8;
}

/// 提供房间相关的服务接口，诸如创建房间、加入房间、添加房间状态监听等。可通过 [NERoomKit.getRoomService] 获取对应的服务实例
abstract class NERoomService {
  /// 创建一个新的房间，只有完成SDK的登录鉴权操作才允许创建房间。创建房间成功后，SDK会拉起房间页面，调用方不用做其他操作
  ///
  /// * [param] 房间参数对象，不能为空
  /// * [opts]  会议选项对象，可空；当未指定时，会使用默认的选项
  /// * [NEResult] 回调接口，该回调不会返回额外的结果数据
  Future<NEResult<JoinRoomInfo>> startRoom(NEStartRoomParams param, NEStartRoomOptions? opts);

  /// 加入一个当前正在进行中的房间，已登录或未登录均可加入房间。加入房间成功后，SDK会拉起房间页面，调用方不用做其他操作
  ///
  /// * [param] 房间参数对象，不能为空
  /// * [NEResult] 回调接口，该回调不会返回额外的结果数据
  Future<NEResult<JoinRoomInfo>> joinRoom(NEJoinRoomParams param, NEJoinRoomOptions? opts);

  /// 离开房前房间
  ///
  /// [closeIfHost] 如果是房间主持人，在离开房间时是否直接结束房间
  Future<NEResult<void>> leaveCurrentRoom(bool closeIfHost);

  /// 获取当前房间详情。如果当前无正在进行中的房间，则回调数据对象为空
  /// [NERoomInfo]
  NERoomInfo? getCurrentRoomInfo();
}
