// Copyright (c) 2014-2021 NetEase, Inc.
// All right reserved.

part of room_kit;

class _RoomUserVideoStreamSubscription {
  final String userId;
  final int streamType;
  final Set<dynamic> owners = {};
  final NERtcVideoRenderer renderer;

  _RoomUserVideoStreamSubscription(this.userId, this.streamType, this.renderer);
}

/// An InheritedWidget which provides [NEInRoomVideoController] instance
class NERoomUserVideoControllerProvider extends InheritedWidget {
  final NEInRoomVideoController videoController;

  NERoomUserVideoControllerProvider({
    Key? key,
    required Widget child,
    required this.videoController,
    int renderCacheSize = 4,
  }) : super(
          key: key,
          child: _RoomUserVideoStreamSubscriberWrapper(
            child: child,
            renderCacheSize: renderCacheSize,
          ),
        );

  static NERoomUserVideoControllerProvider of(BuildContext context) {
    final result = context.dependOnInheritedWidgetOfExactType<
        NERoomUserVideoControllerProvider>();
    assert(result != null,
        'No NERoomUserVideoControllerProvider found in context');
    return result!;
  }

  @override
  bool updateShouldNotify(NERoomUserVideoControllerProvider old) {
    return videoController != old.videoController;
  }
}

class _RoomUserVideoStreamSubscriberWrapper extends StatefulWidget {
  final Widget child;
  final int renderCacheSize;

  const _RoomUserVideoStreamSubscriberWrapper({
    Key? key,
    required this.child,
    this.renderCacheSize = 4,
  }) : super(key: key);

  @override
  __RoomUserVideoStreamSubscriberWrapperState createState() =>
      __RoomUserVideoStreamSubscriberWrapperState();
}

class __RoomUserVideoStreamSubscriberWrapperState
    extends State<_RoomUserVideoStreamSubscriberWrapper> {

  _RoomUserVideoStreamSubscriber? subscriber;

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    subscriber?.dispose();
    subscriber = _RoomUserVideoStreamSubscriber(
        NERoomUserVideoControllerProvider.of(context).videoController,
        widget.renderCacheSize,
    );
  }

  @override
  Widget build(BuildContext context) {
    return _RoomUserVideoStreamSubscriberProvider(
      child: widget.child,
      subscriber: subscriber!,
    );
  }

  @override
  void dispose() {
    subscriber?.dispose();
    super.dispose();
  }
}

class _RoomUserVideoStreamSubscriberProvider extends InheritedWidget {

  final _RoomUserVideoStreamSubscriber subscriber;

  const _RoomUserVideoStreamSubscriberProvider({
    Key? key,
    required Widget child,
    required this.subscriber,
  }) : super(key: key, child: child);

  static _RoomUserVideoStreamSubscriberProvider of(BuildContext context) {
    final result = context.dependOnInheritedWidgetOfExactType<
        _RoomUserVideoStreamSubscriberProvider>();
    assert(result != null,
        'No _RoomUserVideoStreamSubscriberProvider found in context');
    return result!;
  }

  @override
  bool updateShouldNotify(_RoomUserVideoStreamSubscriberProvider old) {
    return subscriber != old.subscriber;
  }
}

/// be responsible for subscribe/unsubscribe video (sub)stream
class _RoomUserVideoStreamSubscriber {

  final NEInRoomVideoController videoController;
  final int renderCacheSize;

  _RoomUserVideoStreamSubscriber(this.videoController, [this.renderCacheSize = 4]);

  final _subStreamRenderCache = <String,NERtcVideoRenderer>{};
  final _renderCache = <String,NERtcVideoRenderer>{};

  // 主流订阅信息
  final _mainStreamSubscriptions = <String,_RoomUserVideoStreamSubscription>{};
  // 辅流订阅信息
  final _subStreamSubscriptions = <String,_RoomUserVideoStreamSubscription>{};

  Future<NERtcVideoRenderer> _dequeueRenderer(dynamic owner, String userId, bool subStream) {
    final subInfo = subStream ? _subStreamSubscriptions : _mainStreamSubscriptions;
    // 复用render
    if (subInfo.containsKey(userId)) {
      subInfo[userId]!.owners.add(owner);
      return Future.sync(() => subInfo[userId]!.renderer);
    }
    if (subStream && _subStreamRenderCache.containsKey(userId)) {
      return Future.sync(() => _subStreamRenderCache.remove(userId) as NERtcVideoRenderer);
    } else if(!subStream && _renderCache.containsKey(userId)){
      return Future.sync(() => _renderCache.remove(userId) as NERtcVideoRenderer);
    }
    return VideoRendererFactory.createVideoRenderer();
  }

  void _subscribeVideoStream(dynamic owner, NERtcVideoRenderer renderer, String userId, int streamType, bool subStream) {
    final subInfo = subStream ? _subStreamSubscriptions : _mainStreamSubscriptions;
    var subscription = subInfo[userId];
    if (subscription == null || subscription.streamType != streamType) {
      Alog.i(
        tag: 'RoomUserVideoStreamSubscriber',
        content: 'subscribeVideoStream: $userId $streamType $subStream',
      );
      final updated = _RoomUserVideoStreamSubscription(userId, streamType, renderer);
      subInfo[userId] = updated;
      updated.owners.addAll(subscription?.owners ?? []);
      subscription = updated;
      if (subStream) {
        videoController.subscribeRemoteVideoSubStream(userId);
      } else {
        videoController.subscribeRemoteVideoStream(userId, streamType);
      }
    }
    if (subStream) {
      videoController.attachRendererToUserVideoSubStream(renderer, userId);
    } else {
      videoController.attachRendererToUserVideoStream(renderer, userId);
    }
    subscription.owners.add(owner);
  }

  void _unsubscribeVideoStream(dynamic owner, NERtcVideoRenderer? renderer, String userId, int streamType, bool subStream) {
    final subInfo = subStream ? _subStreamSubscriptions : _mainStreamSubscriptions;
    final exist = subInfo[userId];
    exist?.owners.remove(owner);
    if (exist != null && exist.owners.isEmpty) {
      Alog.i(
        tag: 'RoomUserVideoStreamSubscriber',
        content: 'unsubscribeVideoStream: $userId $streamType $subStream',
      );
      subInfo.remove(userId);
      if (subStream) {
        videoController.unsubscribeRemoteVideoSubStream(userId);
      } else {
        videoController.unsubscribeRemoteVideoStream(userId);
      }
      if (renderer != null) {
        assert(renderer == exist.renderer);
        if (subStream && _subStreamRenderCache.length < renderCacheSize
            && _subStreamRenderCache[userId] == null) {
          _subStreamRenderCache[userId] = renderer;
        } else if(!subStream && _renderCache.length < renderCacheSize
            && _renderCache[userId] == null) {
          _renderCache[userId] = renderer;
        } else {
          renderer.dispose();
          Alog.i(
            tag: 'RoomUserVideoStreamSubscriber',
            content: 'dispose render: $userId $streamType $subStream',
          );
        }
      }
    }
  }

  void dispose() {
    Alog.i(
      tag: 'RoomUserVideoStreamSubscriber',
      content: 'dispose',
    );
    _subStreamRenderCache.forEach((_,element) {
      element.dispose();
    });
    _renderCache.forEach((_,element) {
      element.dispose();
    });
  }
}

class _UserVideoKey extends LocalKey {
  _UserVideoKey(this.userId, this.subStream);

  final String userId;
  final bool subStream;

  @override
  bool operator ==(Object other) {
    if (other.runtimeType != runtimeType) {
      return false;
    }
    return other is _UserVideoKey &&
        userId == other.userId &&
        subStream == other.subStream;
  }

  @override
  int get hashCode => hashValues(runtimeType, userId, subStream);
}

/// user video view widget
/// it automatically subscribe and unsubscribe video stream for user
class NERoomUserVideoView extends StatefulWidget {

  final NERoomUserVideoViewListener? listener;
  final NERtcVideoViewFitType fitType;
  final Color backgroundColor;
  final bool mirror;
  final String userId;
  final bool subStream;
  final int streamType;

  /// 正常流
  NERoomUserVideoView(
      this.userId, {
        Key? key,
        this.listener,
        this.mirror = false,
        this.streamType = NERtcRemoteVideoStreamType.low,
        this.fitType = NERtcVideoViewFitType.contain,
        this.backgroundColor = const Color(0xFF292933),
      })  : subStream = false,
        super(key: key ?? _UserVideoKey(userId, false));

  /// 辅流
  NERoomUserVideoView.subStream(
      this.userId, {
        Key? key,
        this.listener,
        this.mirror = false,
        this.streamType = NERtcRemoteVideoStreamType.high,
        this.fitType = NERtcVideoViewFitType.contain,
        this.backgroundColor = const Color(0xFF292933),
      })  : subStream = true,
        super(key: key ?? _UserVideoKey(userId, true));

  @override
  _RoomUserVideoViewState createState() => _RoomUserVideoViewState();
}

class _RoomUserVideoViewState extends State<NERoomUserVideoView> with NERtcVideoRendererEventListener {

  NERtcVideoRenderer? renderer;
  late _RoomUserVideoStreamSubscriber subscriber;

  void setupRenderer() {
    final renderer = this.renderer;
    assert(renderer != null, 'setup renderer error');
    if (renderer != null) {
      renderer.setMirror(widget.mirror);
      renderer.rendererEventLister = this;
      subscriber._subscribeVideoStream(this, renderer, widget.userId, widget.streamType, widget.subStream);
    }
  }

  @override
  void initState() {
    super.initState();
    Alog.i(
      tag: 'RoomUserVideoViewState',
      content: 'init for ${widget.userId} ${widget.subStream}',
    );
  }

  @override
  void didUpdateWidget(NERoomUserVideoView oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.mirror != oldWidget.mirror || widget.streamType != oldWidget.streamType) {
      setupRenderer();
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    if (renderer == null) {
      subscriber = _RoomUserVideoStreamSubscriberProvider.of(context).subscriber;
      subscriber._dequeueRenderer(this, widget.userId, widget.subStream).then((value) {
        if (mounted) {
          renderer = value;
          setState(setupRenderer);
        } else {
          subscriber._unsubscribeVideoStream(this, value, widget.userId, widget.streamType, widget.subStream);
        }
      });
    }
  }

  @override
  void dispose() {
    subscriber._unsubscribeVideoStream(this, renderer, widget.userId, widget.streamType, widget.subStream);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final renderer = this.renderer;
    if(renderer == null) return Container();
    return NERtcVideoView(
      renderer,
      fitType: widget.fitType,
      backgroundColor: widget.backgroundColor,
    );
  }

  @override
  void onFirstFrameRendered(int uid) {
    widget.listener?.onFirstFrameRendered(widget.userId);
  }

  @override
  void onFrameResolutionChanged(int uid, int width, int height, int rotation) {
    widget.listener?.onFrameResolutionChanged(widget.userId, width, height, rotation);
  }
}

/// 视频渲染事件监听器
abstract class NERoomUserVideoViewListener {
  void onFrameResolutionChanged(String userId, int width, int height, int rotation);

  void onFirstFrameRendered(String userId);
}
