// Copyright (c) 2014-2020 NetEase, Inc.
// All right reserved.

part of room_service;

/// NIM 相关API服务
class NimApiHelper {
  NimApiHelper._();

  static Future<NEResult<T>> execute<T>(BaseApi api) {
    if (_AuthManager().imNetBroken()) {
      return Future.value(NEResult(code: RoomErrorCode.networkError));
    }
    var completer = Completer<NEResult<T>>();
    _AuthManager()._imLoginCheck().then((code) {
      if (code == RoomErrorCode.success) {
        completer.complete(_ApiHelper._execute(api));
      } else {
        completer
            .complete(Future.value(NEResult(code: RoomErrorCode.networkError)));
      }
    });

    return completer.future;
  }

  static Future<NEResult<dynamic>> sendRequest(NEHttpApiRequest request) {
    return execute(_GenericPassThroughApiRequest(request));
  }

  /// NIM 登陆
  static Future<int> _imLogin(String imAccountId, String imToken, String imAppKey) {
    var completer = Completer<int>();
    NimPlugin().getAuthService().login(AuthInfo(imAccountId, imToken, imAppKey),
        (int code, data) {
      completer.complete(code);
    });
    return completer.future;
  }

  /// NIM登出
  static void _imLogout() => NimPlugin().getAuthService().logout();

  /// 发送普通消息
  static Future<int> sendMessage(IMMessage message) {
    var completer = Completer<int>();
    NimPlugin().getMsgService().sendMessage(
        message: message,
        callback: (int code, data) {
          completer.complete(code);
        });
    return completer.future;
  }

  ///主持人控制协议
  static Future<NEResult<void>> _hostControl(String meetingId, int actionType,
          {List? members}) =>
      execute(_HostControlApi(_HostControlRequest(
          meetingId: meetingId, actionType: actionType, members: members)));

  ///成员发起会控操作
  static Future<NEResult<void>> _memberControl(String meetingId, int actionType,
          {List? avRoomUids}) =>
      execute(_MemberControlApi(_MemberControlRequest(
          meetingId: meetingId,
          actionType: actionType,
          avRoomUids: avRoomUids)));

  /// 主持人关闭会议
  static Future<NEResult<void>> _closeRoom(String meetingId) =>
      execute(_CloseMeetingApi(_CloseMeetingRequest(meetingId: meetingId)));

  /// 获取会议中的成员信息
  /// avRoomUid 不传则表示获取全部成员列表信息
  static Future<NEResult<MeetingInfo>> _getMemberInfo(String meetingId,
          {int? avRoomUid}) =>
      execute(_MeetingInfoApi(
          _MeetingInfoRequest(meetingId: meetingId, avRoomUid: avRoomUid)));

  /// 服务器控制协议
  static Stream<MeetingAction> get _onServerControlReceived =>
      _ServerControlApi().listen();

  /// 在线状态
  static Stream<StatusCode> get _onOnlineStatusChanged =>
      _OnlineStatusApi().listen();

  /// 上传文件
  static Future<NEResult<String>> uploadFile(String filePath) {
    final completer = Completer<NEResult<String>>();
    NimPlugin().getResourceService().uploadAtScene(
        filePath: filePath,
        sceneKey: SceneKey.nimSystemNosScene,
        callback: (int code, String? data) {
          completer.complete(NEResult(code: code, data: data));
        });
    return completer.future;
  }

  /// 开始直播
  static Future<NEResult<NERoomItemLive>> _startLive(
          NERoomItemLive live) =>
      execute(_StartLiveApi(live));

  /// 更新直播
  static Future<NEResult<NERoomItemLive>> _updateLive(
          NERoomItemLive live) =>
      execute(_UpdateLiveApi(live));

  /// 结束直播
  static Future<NEResult<NERoomItemLive>> _stopLive(String meetingId) =>
      execute(_StopLiveApi(meetingId));

  /// 修改昵称
  static Future<NEResult<void>> _modifyNickname(
          String meetingId, String nickname) =>
      execute(_ModifyNicknameApi(
          _ModifyNicknameRequest(meetingId: meetingId, nickname: nickname)));
}
