// Copyright (c) 2014-2020 NetEase, Inc.
// All right reserved.

part of room_service;

class _AuthManager with EventTrackMixin {
  factory _AuthManager() => _instance;

  static final _AuthManager _instance = _AuthManager._internal();

  _AuthManager._internal() : _authState = AuthState.idle;

  int _authState;

  late String _appKey;

  String? _domain;

  /// 是否复用底层IM的登录态
  bool _reuseNIM = false;

  String? _sdkDir;

  _LoginInfo? _loginInfo;

  String? _lastConfigNIMAppKey;

  String? get imAccid => _loginInfo?.imAccid;

  String? get imToken => _loginInfo?.imToken;

  String? get accountId => _loginInfo?.accountId;

  String? get accountToken => _loginInfo?.accountToken;

  String get appKey => _appKey;

  String? get domain => _domain;

  String? get imAppKey => _loginInfo?.imAppKey;

  String? get nrtcAppKey => _loginInfo?.nrtcAppKey;

  String? get personalMeetingId => _loginInfo?.personalMeetingId;

  String? get nickname => _loginInfo?.nickname;

  String? get shortId => _loginInfo?.shortId;

  bool _initialized = false;

  StreamController<int> broadcast = StreamController.broadcast();

  int get state => _authState;

  Future<bool> initialize(
      String appKey, String? domain, bool reuseNIM, String? sdkDir) async {
    _appKey = appKey;
    _domain = domain;
    _reuseNIM = reuseNIM;
    _sdkDir = sdkDir;
    Alog.d(
        tag: _tag,
        moduleName: _moduleName,
        content: 'initialize appKey = $appKey, reuseNIM=$_reuseNIM');
    if (!_initialized) {
      await NimPlugin().config(
        appKey: null,
        options: NIMOptions(sdkDir: _sdkDir),
      ); // 当前不设置，在login的时候在config nim
      _ImAuthManager().authStateChanged().listen((event) {
        if (event == _ImAuthSate.kicked) {
          forceUnauthor(_Strings.accountHasLoginedOnOtherDevice);
        } else if (event == _ImAuthSate.deadlyError) {
          tokenIllegal(_Strings.imFailed);
        } else if (event == _ImAuthSate.done) {
          SDKConfig.reFetch();
        }
      });
      _initialized = true;
    }
    return true;
  }

  void updateAppKey(String? appKey) {
    Alog.d(
        tag: _tag,
        moduleName: _moduleName,
        content: 'updateAppKey appKey = $appKey');
    if (TextUtils.isEmpty(appKey)) {
      return;
    }
    if (_appKey != appKey) {
      _appKey = appKey!;
      SDKConfig.reFetch(forceUpdate: true);
    } else {
      SDKConfig.reFetch();
    }
  }

  Future<NEResult<void>> loginWithNEMeeting(
      String username, String password) async {
    if (TextUtils.isEmpty(username)) {
      return _loginError(RoomErrorCode.paramsError, 'username is null');
    }

    if (TextUtils.isEmpty(password)) {
      return _loginError(RoomErrorCode.paramsError, 'password is null');
    }

    if (TextUtils.isEmpty(appKey)) {
      return _loginError(RoomErrorCode.paramsError, 'appKey is null');
    }

    trackPeriodicEvent(TrackEventName.loginMeetingServer);
    var response = await HttpApiHelper._loginWithNEMeeting(username, password);
    return handleLoginErrorWithCode(response);
  }

  Future<NEResult<void>> loginWithToken(
      String? accountId, String? accountToken) async {
    if (TextUtils.isEmpty(accountId) && TextUtils.isEmpty(accountToken)) {
      return _loginError(RoomErrorCode.paramsError, 'accountToken is null');
    }

    trackPeriodicEvent(TrackEventName.loginMeetingServer);
    var response = await HttpApiHelper._loginWithToken(accountId!, accountToken!);
    return handleLoginErrorWithCode(response);
  }

  Future<NEResult<void>> loginWithSSOToken(String ssoToken) async {
    if (TextUtils.isEmpty(ssoToken)) {
      return _loginError(RoomErrorCode.paramsError, 'ssoToken is null');
    }

    trackPeriodicEvent(TrackEventName.loginMeetingServer);
    var response = await HttpApiHelper._loginWithSSOToken(ssoToken);
    return handleLoginErrorWithCode(response);
  }

  Future<NEResult<void>> handleLoginErrorWithCode(
      NEResult<_LoginResponse> response) async {
    if (response.code == RoomErrorCode.success) {
      return await _imLoginCheckWithLogin(loginResponse: response.data);
    } else {
      if (response.code == RoomErrorCode.verifyError ||
          response.code == RoomErrorCode.tokenError ||
          response.code == RoomErrorCode.passwordError ||
          response.code == RoomErrorCode.loginPasswordError) {
        trackPeriodicEvent(TrackEventName.loginMeetingServerFailed,
            extra: {'value': response.code});
        logout();
      }
    }
    return response;
  }

  Future<NEResult<void>> autoLoginWithLocalToken() async {
    var _cacheAppKey = await _SDKPreferences.appKey;
    if (_cacheAppKey != appKey) {
      return _loginError(RoomErrorCode.paramsError, 'appKey not same');
    }
    var _cachedAccountId = await _SDKPreferences.accountId;
    var _cachedAccountToken = await _SDKPreferences.accountToken;
    return loginWithToken(_cachedAccountId, _cachedAccountToken);
  }

  NEResult _loginError(int code, String msg) {
    Alog.i(
        tag: _tag,
        moduleName: _moduleName,
        content: 'AuthManager: login error: $code#$msg');
    updateState(state: AuthState.idle);
    return NEResult(code: code, msg: msg);
  }

  void _loginSuccess(_LoginInfo _loginInfo) {
    Alog.i(
        tag: _tag,
        moduleName: _moduleName,
        content: 'LoginInfo: ${_loginInfo.toString()}');
    this._loginInfo = _loginInfo;
    trackPeriodicEvent(TrackEventName.loginMeetingServerSuccess);
    updateAppKey(_loginInfo.appKey);
    updateState(state: AuthState.authorized);
    if (UserProfile.isAnonymous) {
      _SDKPreferences.setAccountToken('');
      _SDKPreferences.setAccountId('');
    } else {
      _SDKPreferences.setAccountToken(_loginInfo.accountToken);
      _SDKPreferences.setAccountId(_loginInfo.accountId);
    }
    _SDKPreferences.setAppKey(_loginInfo.appKey);
    Alog.i(
        tag: _tag,
        moduleName: _moduleName,
        content:
            'AuthManager: login done, anonymous=${UserProfile.isAnonymous}');
  }

  bool imNetBroken() {
    return _ImAuthManager().netBroken;
  }

  Future<NEResult> _imLoginCheckWithLogin(
      {_LoginResponse? loginResponse}) async {
    /// 复用IM的情况下，不支持匿名登录
    if (UserProfile.isAnonymous && _reuseNIM) {
      return NEResult(
          code: RoomErrorCode.loginErrorAnonymousLoginNotSupport,
          msg: _Strings.anonymousLoginNotSupportedWhenReuseNIM);
    }

    var info = loginResponse2LoginInfo(loginResponse);
    Alog.i(
        tag: _tag,
        moduleName: _moduleName,
        content: 'loginInfo ====== ${info?.toString()}');
    if (info == null) {
      return NEResult(
          code: RoomErrorCode.paramsError, msg: _Strings.serverParamError);
    }
    int code;
    String msg;

    /// 处理IM复用逻辑
    var authInfo = await NimPlugin().getAuthService().getLoginInfo();
    var isIMLogin = authInfo != null && authInfo.isValid;
    if (_reuseNIM) {
      if (!isIMLogin) {
        /// 复用IM，但IM未登录 => 登录失败
        code = RoomErrorCode.loginErrorIMNotLogin;
        msg = _Strings.imIsNotLoginWhenReuseNIM;
      } else if (!(authInfo?.isMatch(
              info.imAccid, info.imToken, info.imAppKey) ??
          false)) {
        /// 复用IM，但IM登录账号与会议账号不匹配 => 登录失败
        code = RoomErrorCode.loginErrorIMAccountNotMatch;
        msg = _Strings.accountUnMatchWhenReuseNIM;
        logout();
      } else {
        code = RoomErrorCode.success;
        msg = 'SUCCESS';
        _loginSuccess(info);

        /// 复用IM，不会进行IM的登录，需要主动监听IM的状态
        _ImAuthManager().startAuthWatch();
      }
      return NEResult(code: code, msg: msg);
    } else {
      updateState(state: AuthState.authorizing);

      /// Config NIM SDK Here If NIM AppKey Changed
      /// IOS 需要
      if (!TextUtils.nonEmptyEquals(_lastConfigNIMAppKey, info.imAppKey)) {
        await NimPlugin().config(
          appKey: info.imAppKey,
          authInfo: null,
          options: NIMOptions(sdkDir: _sdkDir),
        );
        _lastConfigNIMAppKey = info.imAppKey;
      }
      return _ImAuthManager()
          .login(info.imAccid, info.imToken, info.imAppKey)
          .then((code) {
        if (code == RoomErrorCode.success) _loginSuccess(info);
        return NEResult(code: code);
      });
    }
  }

  Future<int> _imLoginCheck() async {
    var code = 0;
    var info = _loginInfo;

    /// 处理IM复用逻辑
    var authInfo = await NimPlugin().getAuthService().getLoginInfo();
    var isIMLogin = authInfo != null && authInfo.isValid;
    if (_reuseNIM) {
      if (!isIMLogin) {
        /// 复用IM，但IM未登录 => 登录失败
        code = RoomErrorCode.loginErrorIMNotLogin;
      } else if (!(authInfo?.isMatch(
              info?.imAccid, info?.imToken, info?.imAppKey) ??
          false)) {
        /// 复用IM，但IM登录账号与会议账号不匹配 => 登录失败
        code = RoomErrorCode.loginErrorIMAccountNotMatch;
      } else {
        code = RoomErrorCode.success;
      }
      return code;
    } else {
      if (_authState == AuthState.authorized && isIMLogin) {
        return RoomErrorCode.success;
      }
      return _ImAuthManager()
          .login(info?.imAccid, info?.imToken, info?.imAppKey)
          .then((code) {
        return code;
      });
    }
  }

  void logout() {
    updateState(state: AuthState.unAuthorizing);
    _loginInfo = null;
    SDKConfig.initialized = false;
    updateState(state: AuthState.idle);

    /// 不复用IM时才允许操作IM退登
    if (!_reuseNIM) {
      _ImAuthManager().logout();
    } else {
      _ImAuthManager().stopAuthWatch();
    }
    _SDKPreferences.setAccountToken('');
    _SDKPreferences.setAccountId('');
    _SDKPreferences.setAppKey('');
    Alog.i(tag: _tag, moduleName: _moduleName, content: 'AuthManager: logout');
  }

  void tokenIllegal(String errorTip) {
    updateState(state: AuthState.tokenExpired, errorTips: errorTip);
    logout();
  }

  void forceUnauthor(String errorTip) {
    updateState(state: AuthState.kickOut, errorTips: errorTip);
    logout();
  }

  void updateState({required int state, errorTips = ''}) {
    Alog.i(
        tag: _tag,
        moduleName: _moduleName,
        content: 'AuthManager: update state=${AuthState.stringify(state)}');
    _authState = state;

    /// notify
    broadcast.add(_authState);
  }

  Stream<int> obAuthState() {
    return broadcast.stream;
  }

  _LoginInfo? loginResponse2LoginInfo(_LoginResponse? loginResponse) {
    if (loginResponse == null) {
      return null;
    }
    return _LoginInfo(
        imAccid: loginResponse.imAccid,
        imToken: loginResponse.imToken,
        accountId: loginResponse.accountId,
        accountToken: loginResponse.accountToken,
        appKey: loginResponse.appKey,
        imAppKey: loginResponse.imAppKey,
        nrtcAppKey: loginResponse.nrtcAppKey,
        personalMeetingId: loginResponse.personalMeetingId,
        nickname: loginResponse.nickname,
        shortId: loginResponse.shortId);
  }
}

class _LoginInfo {
  final String imAccid;

  final String imToken;

  final String accountId;

  final String accountToken;

  final String appKey;

  final String imAppKey;

  final String nrtcAppKey;

  final String? personalMeetingId;

  final String nickname;

  final String? shortId;

  _LoginInfo(
      {required this.imAccid,
      required this.imToken,
      required this.accountId,
      required this.accountToken,
      required this.appKey,
      required this.imAppKey,
      required this.nrtcAppKey,
      required this.personalMeetingId,
      required this.nickname,
      this.shortId});

  @override
  bool operator ==(other) {
    // 判断是否是非
    if (other is! _LoginInfo) {
      return false;
    }
    final loginInfo = other;
    return imAccid == loginInfo.imAccid &&
        imToken == loginInfo.imToken &&
        accountId == loginInfo.accountId &&
        accountToken == loginInfo.accountToken &&
        appKey == loginInfo.appKey &&
        imAppKey == loginInfo.imAppKey &&
        nrtcAppKey == loginInfo.nrtcAppKey &&
        personalMeetingId == loginInfo.personalMeetingId &&
        nickname == loginInfo.nickname &&
        shortId == loginInfo.shortId;
  }

  @override
  int get hashCode =>
      imAccid.hashCode ^
      imToken.hashCode ^
      accountId.hashCode ^
      accountToken.hashCode ^
      appKey.hashCode ^
      imAppKey.hashCode ^
      nrtcAppKey.hashCode ^
      personalMeetingId.hashCode ^
      nickname.hashCode ^
      shortId.hashCode;

  @override
  String toString() {
    return 'LoginInfo{imAccid: $imAccid, '
        'accountId: $accountId, personalMeetingId: $personalMeetingId, '
        'nickname: $nickname, shortId: $shortId,'
        ' appKey: $appKey';
  }
}
