// Copyright (c) 2014-2020 NetEase, Inc.
// All right reserved.

part of room_service;

class _ImAuthSate {
  /// init state
  static const int init = 0;

  /// 网络错误之类的，可重试
  static const int unlogin = 1;

  /// login state
  static const int logining = 2;

  /// done
  static const int done = 3;

  /// kicked，退到首页
  static const int kicked = 4;

  /// 没有办法再连上的，退到首页
  static const int deadlyError = 5;
}

class _ImAuthManager with EventTrackMixin {
  final Connectivity _connectivity = Connectivity();

  StreamSubscription<ConnectivityResult>? _connectivitySubscription;

  StreamSubscription<StatusCode>? _imAuthSubscription;

  late StreamController<int> imAuthNotify;

  ConnectivityResult? _result;

  int authState = _ImAuthSate.init;

  bool _netBroken = false;

  List<int> imErrorCode = [302, 404, 414, 317, 403, 422, 417];

  static final _ImAuthManager _singleton = _ImAuthManager._internal();

  factory _ImAuthManager() => _singleton;

  bool isNimInitialized = false;

  _ImAuthManager._internal() {
    imAuthNotify = StreamController.broadcast();
  }

  // Platform messages are asynchronous, so we initialize in an async method.
  Future<void> initConnectivity() async {
    // Platform messages may fail, so we use a try/catch PlatformException.
    try {
      _result = await _connectivity.checkConnectivity();
    } on PlatformException catch (e) {
      print(e.toString());
    }
  }

  void _updateConnectionStatus(ConnectivityResult result) {
    Alog.i(
        tag: _tag,
        moduleName: _moduleName,
        content: '_ImAuthManager: networkChanged(from=$_result, to=$result)');
    if (_result != result && (result != ConnectivityResult.none)) {
      _result = result;

      /// reLogin
      /// login(_AuthManager().imAccid, _AuthManager().imToken, _AuthManager().imAppKey);
    }
  }

  Future<int> login(String? imAccid, String? imToken, String? imAppKey) async {
    Alog.i(
        tag: _tag,
        moduleName: _moduleName,
        content: '_ImAuthManager: login $authState');
    // 检查参数
    if (TextUtils.isEmpty(imAccid) || TextUtils.isEmpty(imToken)) {
      trackPeriodicEvent(TrackEventName.loginImFailed,
          extra: {'value': RoomErrorCode.paramsError});
      return Future.value(RoomErrorCode.paramsError);
    }

    // 如果是已成功状态 直接返回
    if (authState == _ImAuthSate.done) {
      return Future.value(RoomErrorCode.success);
    }

    trackPeriodicEvent(TrackEventName.loginIm);

    var ret = await _login(imAccid, imToken, imAppKey);
    var completer = Completer<int>();

    if (ret == -200) {
      // 正在登录中
      StreamSubscription<int>? ss;
      ss = authStateChanged().listen((event) {
        if (event == _ImAuthSate.done) {
          // 状态迁移到done，说明登录成功了
          ss?.cancel();
          completer.complete(RoomErrorCode.success);
          trackPeriodicEvent(TrackEventName.loginImSuccess);
        } else if (event == _ImAuthSate.unlogin) {
          // 状态迁移到init，说明登录失败了
          ss?.cancel();
          trackPeriodicEvent(TrackEventName.loginImFailed,
              extra: {'value': _ImAuthSate.unlogin});
          completer.complete(RoomErrorCode.failed);
        } else if (event == _ImAuthSate.init) {
          // 状态迁移到init，说明登录失败了
          ss?.cancel();
          trackPeriodicEvent(TrackEventName.loginImFailed,
              extra: {'value': _ImAuthSate.init});
          completer.complete(RoomErrorCode.failed);
        } else {
          ss?.cancel();
        }
      });
    } else {
      if (ret == RoomErrorCode.success) {
        trackPeriodicEvent(TrackEventName.loginImSuccess);
      } else {
        trackPeriodicEvent(TrackEventName.loginImFailed, extra: {'value': ret});
      }
      completer.complete(ret);
    }

    return completer.future;
  }

  Future<int> _login(String? imAccid, String? imToken, String? imAppKey) async {
    /// avoid duplicate login
    if (authState == _ImAuthSate.logining) {
      return Future.value(-200);
    }

    if (authState == _ImAuthSate.done) {
      return Future.value(RoomErrorCode.success);
    }

    authState = _ImAuthSate.logining;
    imAuthNotify.add(authState);

    /// await startWatchNet();
    startAuthWatch();

    // call im api
    var code = await NimApiHelper._imLogin(imAccid!, imToken!, imAppKey!);
    return Future.value(code);
  }

  void logout() {
    _logoutInternal();
    authState = _ImAuthSate.init;
    imAuthNotify.add(authState);
  }

  void _logoutInternal() {
    /// stopNetWatch();
    stopAuthWatch();
    // call im api
    NimApiHelper._imLogout();
  }

  void stopNetWatch() {
    _connectivitySubscription?.cancel();
    _connectivitySubscription = null;
  }

  /// 切换账号重装watch
  void startWatchNet() async {
    if (_connectivitySubscription == null) {
      //获取一次最新的网络状态
      await initConnectivity();
      _connectivitySubscription =
          _connectivity.onConnectivityChanged.listen(_updateConnectionStatus);
    }
  }

  void stopAuthWatch() {
    _imAuthSubscription?.cancel();
    _imAuthSubscription = null;
    _netBroken = false;
  }

  /// when listen first receive unLogin
  void startAuthWatch() {
    stopAuthWatch();
    _imAuthSubscription = NimApiHelper._onOnlineStatusChanged.listen((event) {
      Alog.i(
          tag: _tag,
          moduleName: _moduleName,
          content: '_ImAuthManager: onEvent($event)');
      if (event == StatusCode.kickOut ||
          event == StatusCode.kickByOtherClient) {
        /// 被踢
        Alog.w(
            tag: _tag,
            moduleName: _moduleName,
            content: '_ImAuthManager: kick now!');
        authState = _ImAuthSate.kicked;
      } else if (event == StatusCode.unLogin) {
        /// 等会重试
        authState = _ImAuthSate.unlogin;
      } else if (event == StatusCode.forbidden ||
          event == StatusCode.verError ||
          event == StatusCode.pwdError) {
        /// 不可恢复的
        Alog.w(
            tag: _tag,
            moduleName: _moduleName,
            content: '_ImAuthManager: deadly error now!');
        authState = _ImAuthSate.deadlyError;
      } else if (event == StatusCode.logining) {
        /// 正在登录
        authState = _ImAuthSate.logining;
      } else if (event == StatusCode.logined) {
        /// 登录成功
        authState = _ImAuthSate.done;
      } else {
        /// 其他情况
        authState = _ImAuthSate.init;
      }
      imAuthNotify.add(authState);

      _netBroken = (event == StatusCode.netBroken);
    });
  }

  bool get netBroken => _netBroken;

  Stream<int> authStateChanged() => imAuthNotify.stream;
}
